/*
 * Routine conu par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <string.h>
#include <ctype.h>
#include <math.h>

#include "lf32.h"
#include "op_fftp.h"

/* Evaluation de t=(a*b+c)
 * et retour de t%d et de q=t/d.
 * Faite par Fabrice Bellard
 */
 long mul_add_div( long *q1,
				      long  a ,
				      long  b ,
				     long  c ,
				     long  d)
{
 unsigned long q,r;
 asm(" movl %2,%%eax ; mull %3 ; addl %4,%%eax ; adcl $0,%%edx ; divl %5 ; movl %%eax,%0 ; movl %%edx,%1 "
	: "=g" (q), "=g" (r)
	: "g" (a), "g" (b), "g" (c), "g" (d)
	: "%eax", "%edx"
	);
 *q1=q;
 return r;
}

 long mul_add_div_ul( unsigned long *q1,
			    unsigned long  a ,
			    unsigned long  b ,
			    unsigned long  c ,
			    unsigned long  d)
{
 unsigned long q,r;
 asm(" movl %2,%%eax ; mull %3 ; addl %4,%%eax ; adcl $0,%%edx ; divl %5 ; movl %%eax,%0 ; movl %%edx,%1 "
	: "=g" (q), "=g" (r)
	: "g" (a), "g" (b), "g" (c), "g" (d)
	: "%eax", "%edx"
	);
 *q1=q;
 return r;
}

/* Evaluation de t=(a*b+c)
 * et retour de t%2^30 et de q=t/2^30 .
 * Faite par Raphal Pasquier, adpat d'une routine de Fabrice Bellard
 */

/*inline*/
unsigned long mul_add_div_normalized_bin(unsigned long *q1,
				         unsigned long  a ,
				         unsigned long  b ,
				         unsigned long  c )
{
	 unsigned long qx , rx ;
	 asm("  movl %2,%%eax ; mull %3 ; addl %4,%%eax ; adcl $0, %%edx ; sall $2, %%edx ; movl %%eax, %%ecx ; shr $30, %%eax ; addl %%eax, %%edx ; and $0x3FFFFFFF, %%ecx ; movl %%edx,%0 ; movl %%ecx,%1 "
			 : "=g" (qx), "=g" (rx)
			 : "g" (a), "g" (b), "g" (c)
			 : "%eax", "%edx"
			 );
   *q1=qx;
	 return rx ;
}


/* Evaluation de t=(a*BASE32_BIN+b)
 * et retour de t%d et de q=t/d .
 * Faite par Raphal Pasquier, adpat d'une routine de Fabrice Bellard
 */

/*inline*/
unsigned long mulB_add_div_ul(unsigned long *q,
				         unsigned long  a ,
				         unsigned long  b ,
				         unsigned long  d )
{
	 unsigned long qx , rx ;
	 asm("  movl %2,%%eax ; movl %2,%%edx ; shr $2,%%edx ; sall $30,%%eax ; addl %3,%%eax ; adcl $0,%%edx ; divl %4 ; movl %%edx,%1 ; movl %%eax,%0"
			 : "=g" (qx), "=g" (rx)
			 : "g" (a), "g" (b), "g" (d)
			 : "%eax", "%edx"
			 );
   *q=qx;
	 return rx ;
}

void createLi( Li* li , long nb )
{
 long i ;

 li->signe = 0 ;
 li->taille_mantisse = nb ;
 li->mantisse = (long*) calloc( nb , sizeof(long) ) ;
 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;
}

void destroyLi( Li* li )
{
 long i,nb=li->taille_mantisse ;

 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;
 free( li->mantisse ) ;
 li->taille_mantisse = -10 ;
 li->signe = -2 ;
}

long indiceMax( long* chiffre , Li* a )
{
 long i , taille ;

 if ( a->signe == 0 )
	{
	 *chiffre = 0 ;
	 return 0 ;
	}
 taille = a->taille_mantisse ;
 for ( i = taille-1 ; i >= 0 ; i-- )
	if ( a->mantisse[i] != 0 ) break ;
 if ( i < 0 ) i = 0 ;
 *chiffre = a->mantisse[i] ;
 return i ;
}

long div_R_deci( Li* li , long D )
// necessary 0 < D < Base32 where Base32 = 10^9 and li positive.
// return the rest of the division.
{
 long i , r , q , n ;
 unsigned long tempon ;

 if ( li->signe == 0 ) return 0 ;
 r = 0 ;
 n = li->taille_mantisse - 1 ;
 for( i = n ; i >= 0 ; --i )
        {
	 //a = li->Mantisse[i] + r * 10000 ;
	 //q = a / D ;
	 //r = a - q * D ;
	 tempon = li->mantisse[i] ;
	 r = (long) mul_add_div( &q , r , BASE32 , tempon , D ) ;
	 li->mantisse[i] = q ;
        }
 for( i = 0 ; i <= n ; ++i )
        if ( li->mantisse[i] != 0 ) break ;
 if ( i > n ) li->signe = 0 ;
 return r ;
}

long divB_deci( Li* res , Li* li , long D )
// Je suppose que 0 < diviseur < BASE32 = 10^9 and li positive.
// res <- a / diviseur
// return the rest of the division.
{
 long i , r , q , n ;
 unsigned long tempon ;

 if ( li->signe == 0 )
	{
	 res->signe = 0 ;
	 return 0 ;
	}
 r = 0 ;
 n = li->taille_mantisse - 1 ;

 for( i = n ; i >= 0 ; i-- )
	{
	 tempon = li->mantisse[i] ;
	 r = (long) mul_add_div( &q , r , BASE32 , tempon , D ) ;
	 res->mantisse[i] = q ;
	}

 for( i = 0 ; i <= n ; ++i )
        if ( res->mantisse[i] != 0 ) break ;
 if ( i > n ) res->signe = 0 ; else res->signe = 1 ;
 return r ;
}

int create_BigFloat_deci(Li* bf , char* BigFloat )
// Cette routine engendre un rel long  partir de la chane de caractres BigFloat.
// et sctocke le rsultat dans bf que je suppose pas encore crer.
// bf a cette structure :
// *- bf a deux chiffres pour la partie entire  la fin du vecteur
// *- bf.Mantisse[0]/B^(n-2)+bf.Mantisse[1]/B^(n-3)+....+bf.Mantisse[n-2]+B*bf.Mantisse[n-1]
// Si elle n'y arrive pas, elle renvoie 1 sinon 0 .
{
 long i , val , pe , c , n = 0 , sign ;
 Li secours ;

 if ( strlen( BigFloat ) == 0 )
	{
	 bf->signe = 2 ; // problme
	 bf->mantisse = NULL ;
	 return 1 ;
	}

 sign = 1 ;
 if ( *BigFloat == '-' )
        {
         sign = -1 ;
         BigFloat++ ;
        }
 if ( *BigFloat == '+' ) BigFloat++ ;
 pe = 0 ;

 while ((*BigFloat != 0) && (*BigFloat !='.') && (*BigFloat !=',') && (*BigFloat !='e') && (*BigFloat != 'E'))
        {
	 if (isdigit(*BigFloat) == 0)
	 	{
		 bf->signe = 2 ;
		 bf->mantisse = NULL ;
		 return 1 ; // problme
		}
	 val = (*BigFloat) - '0' ;
  	 pe = 10 * pe + val ;
    	 BigFloat++;
        }

 n = strlen( BigFloat ) / 4 + 3 ;
 createLi( bf , n ) ;
 bf->signe = sign ;

 bf->mantisse[n-2] = pe ;
 if ( *BigFloat == 0 )
 	{
	 if ( pe == 0 ) bf->signe = 0 ;
 	 return 0 ; // c'est un nombre entier
	}

 if ( ( *BigFloat == '.' ) || ( *BigFloat == ',' ) )
        {
         BigFloat++ ;
         c = 0 ;
         pe = 0 ;
         i = n-3 ;
         while (( i >= 0 ) && ( *BigFloat != 0 ) )
                {
		 if ((*BigFloat == 'e') || (*BigFloat == 'E')) break ;
		 if (isdigit(*BigFloat) == 0)
		 	{
			 bf->signe = 2 ;
			 return 1 ; // problme
			}
		 val = (*BigFloat) - '0' ;
     		 pe = 10 * pe + val ;
                 c++ ;
                 if ( c == 9 )  // car la base est 10^9
                        {
                         c = 0 ;
                         bf->mantisse[i] = pe ;
                         i-- ;
                         pe = 0 ;
                        }
                 BigFloat++ ;
                }
         if (( i >= 0 ) && ( c > 0))
                {
                 switch (c)
                        {
                         case 1 : pe *= 100000000 ; break ;
                         case 2 : pe *= 10000000  ; break ;
                         case 3 : pe *= 1000000   ; break ;
                         case 4 : pe *= 100000    ; break ;
                         case 5 : pe *= 10000     ; break ;
                         case 6 : pe *= 1000      ; break ;
                         case 7 : pe *= 100       ; break ;
                         case 8 : pe *= 10        ; break ;
                        }
                 bf->mantisse[i] = pe ;
                }
        }

 if (( *BigFloat =='e') || ( *BigFloat =='E'))
	{
	 BigFloat++ ;
	 if ( *BigFloat != '-' )
		{
		 bf->signe = 2 ;
		 return 1 ; //problme
		}
	 BigFloat++ ;
	 pe = 0 ;
	 while ( *BigFloat != 0)
		{
		 if (isdigit(*BigFloat) == 0)
			{
			 bf->signe = 2 ;
			 return 1 ; // problme
			}
		 val = (*BigFloat) - '0' ;
		 pe = 10 * pe + val ;
		 BigFloat++;
		}
	 n = bf->taille_mantisse + pe/9 ;
	 createLi( &secours, n ) ;
	 secours.signe = bf->signe ;
	 for( i = 0 ; i < bf->taille_mantisse ; i++ )
		secours.mantisse[i+(pe/9)] = bf->mantisse[i] ;
	 destroyLi( bf ) ;
	 createLi( bf , secours.taille_mantisse ) ;
	 bf->signe = secours.signe ;
	 for( i = 0 ; i < secours.taille_mantisse ; i++ )
		bf->mantisse[i] = secours.mantisse[i] ;
	 destroyLi( &secours ) ;
	 for( i = 0 ; i < pe ; i++ )
	 	{
		 div_R_deci( bf , 10 ) ;
		}
	}
	else
	{
	 if ( *BigFloat != 0 )
		{
		 bf->signe = 2 ;
		 return 1 ; //problme
		}
	}

 for( i = n - 1 ; i >= 0 ; i-- ) if ( bf->mantisse[i]  != 0 ) break ;
 if ( i < 0 ) bf->signe = 0 ;

 return 0 ;
}

void printLi( Li* li )
{
  long i , n ;

 n = li->taille_mantisse ;
 printf(" Precision = %ld\n" , n ) ;
 printf(" Signe = %i\n" , li->signe ) ;
 for( i = n - 1 ; i >= 0 ; i-- ) printf("%ld " , li->mantisse[i] ) ;
 printf("\n") ;
}

int create_BigFloat_deci2(Li* bf , long taille , char* BigFloat )
// Cette routine engendre un rel long  partir de la chane de caractres BigFloat.
// et sctocke le rsultat dans bf que je suppose pas encore crer.
// bf a cette structure :
// *- bf a deux chiffres pour la partie entire  la fin du vecteur
// *- bf.Mantisse[0]/B^(n-2)+bf.Mantisse[1]/B^(n-3)+....+bf.Mantisse[n-2]+B*bf.Mantisse[n-1]
// Si elle n'y arrive pas, elle renvoie 1 sinon 0 .
{
 long i , val , pe , c , sign , expo=0 , signExpo=1 , taille_temp=0 ;
 Li temp ;
 char* p ;

 if ( strlen( BigFloat ) == 0 )
	{
	 bf->signe = 2 ; // problme
	 bf->mantisse = NULL ;
	 return 1 ;
	}
 createLi( bf , taille ) ;

 sign = 1 ;
 p = BigFloat ;
 while ( *p != '\0' )
	{
	 if (( *p =='e') || ( *p =='E')) break ;
	 p++ ;
	}
 if (( *p =='e') || ( *p =='E'))
	{
	 p++ ;
	 signExpo = 1 ;
	 if ( *p == '-' )
		{
		 signExpo = -1 ;
		 p++ ;
		}
	 if ( *p == '+' )
		{
		 signExpo = 1 ;
		 p++ ;
		}
	 pe = 0 ;
	 while ( *p != 0)
		{
		 if (isdigit(*p) == 0)
			{
			 destroyLi( bf ) ;
		 	 bf->mantisse = NULL ;
			 return 1 ; // problme
			}
		 val = (*p) - '0' ;
		 pe = 10 * pe + val ;
		 p++;
		}
	 expo = pe ;
	}
	else expo = 0 ;

 taille_temp = taille + 1 + expo / 9 ;
 createLi( &temp , taille_temp ) ;

 while (*BigFloat == ' ') BigFloat++ ;
 if ( *BigFloat == '-' )
        {
         sign = -1 ;
         BigFloat++ ;
        }
 if ( *BigFloat == '+' ) BigFloat++ ;
 temp.signe = sign ;
 pe = 0 ; // partie entire du rel.

 while ((*BigFloat != 0) && (*BigFloat !='.') && (*BigFloat !=',') && (*BigFloat !='e') && (*BigFloat != 'E'))
        {
	 if (isdigit(*BigFloat) == 0)
	 	{
		 bf->signe = 2 ;
		 destroyLi( bf ) ;
		 destroyLi( &temp ) ;
		 bf->mantisse = NULL ;
		 return 1 ; // problme
		}
	 val = (*BigFloat) - '0' ;
  	 pe = 10 * pe + val ;
    	 BigFloat++;
        }

 temp.mantisse[taille_temp-1] = pe / BASE32 ;
 temp.mantisse[taille_temp-2] = pe % BASE32 ;

 if ( *BigFloat == 0 )
 	{
	 if ( pe == 0 ) bf->signe = 0 ;
	 destroyLi( &temp ) ;
 	 return 0 ; // c'est un nombre entier
	}

 if ( ( *BigFloat == '.' ) || ( *BigFloat == ',' ) )
        {
         BigFloat++ ;
         c = 0 ;
         pe = 0 ;
         i = taille_temp-3 ;
         while (( i >= 0 ) && ( *BigFloat != 0 ) )
                {
		 if ((*BigFloat == 'e') || (*BigFloat == 'E') ||
		     (*BigFloat == '\0') || (*BigFloat == '\n')) break ;
		 if (isdigit(*BigFloat) == 0)
		 	{
			 bf->signe = 2 ;
		 	 destroyLi( bf ) ;
		 	 bf->mantisse = NULL ;
			 destroyLi( &temp ) ;
			 return 1 ; // problme
			}
		 val = (*BigFloat) - '0' ;
     		 pe = 10 * pe + val ;
                 c++ ;
                 if ( c == 9 )  // car la base est 10^9
                        {
                         c = 0 ;
                         temp.mantisse[i] = pe ;
                         i-- ;
                         pe = 0 ;
                        }
                 BigFloat++ ;
                }
         if (( i >= 0 ) && ( c > 0))
                {
                 switch (c)
                        {
                         case 1 : pe *= 100000000 ; break ;
                         case 2 : pe *= 10000000  ; break ;
                         case 3 : pe *= 1000000   ; break ;
                         case 4 : pe *= 100000    ; break ;
                         case 5 : pe *= 10000     ; break ;
                         case 6 : pe *= 1000      ; break ;
                         case 7 : pe *= 100       ; break ;
                         case 8 : pe *= 10        ; break ;
                        }
                 temp.mantisse[i] = pe ;
                }
        }
 if (signExpo == -1) for( i = 0 ; i < expo ; i++ ) div_R_deci( &temp , 10 ) ;
	         else    for( i = 0 ; i < expo ; i++ ) mulB2_R_deci( &temp , 10 ) ;

 for( i = taille-1 ; i >= 0 ; i--) bf->mantisse[i] = temp.mantisse[taille_temp+i-taille] ;
 bf->signe = temp.signe ;
 for( i = taille - 1 ; i >= 0 ; i-- ) if ( bf->mantisse[i]  != 0 ) break ;
 if ( i < 0 ) bf->signe = 0 ;

 destroyLi( &temp ) ;
 return 0 ;
}


void copy( Li* nouveau , Li* li )
// Je suppose que li et nouveau ont mme taille.
{
 long i , n ;

 nouveau->signe = li->signe ;
 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void copyAndCreate( Li* nouveau , Li* li )
{
 long i , n ;

 nouveau->signe = li->signe ;
 n = li->taille_mantisse ;
 nouveau->taille_mantisse = n ;
 nouveau->mantisse = (long*) calloc( n , sizeof(long) ) ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void permutLi( Li* li2 , Li* li )
{
 long i , a ;

 a = li2->signe ;
 li2->signe = li->signe ;
 li->signe = a ;
 for( i = 0 ; i < li->taille_mantisse ; i++ )
 	{
	 a = li2->mantisse[i] ;
	 li2->mantisse[i] = li->mantisse[i] ;
	 li->mantisse[i] = a ;
	}
}

void mulB_deci( Li* prod , Li* li , long m )
// multiplication par un chiffre.
// je ne gre pas le dbordement.
{
 long i , n , q ;
 unsigned long tempon ;

 n = li->taille_mantisse ;
 if (( li->signe == 0 ) || ( m == 0 ))
        {
         prod->signe = 0 ;
         for(i = 0 ; i < n ; i++ ) prod->mantisse[i]  = 0 ;
         return ;
        }
 if ( m > 0 ) prod->signe = li->signe ;
	else
	{
	 prod->signe = -li->signe ;
	 m = -m ;
	}
 q = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
	 tempon = li->mantisse[i] ;
	 prod->mantisse[i] = mul_add_div( &q , m , tempon , q , BASE32 ) ;
        }
}

long mulB2_R_deci( Li* al , long m  )
// Je suppose que m est positif et qu'il reprsente un entier
// plus grand ( priori que BASE32=10^9) mais qui loge dans un long.
//  0 <= m <= 2^31 -1
// la routine renvoie 1 s'il y a un dbordement et 0 sinon.
{
 unsigned long i , carry=0 , n ;
 unsigned long tempon ;

 n = al->taille_mantisse ;
 if ( al->signe == 0 ) return 0 ;
 if ( m == 0 )
        {
         for( i = 0 ; i < n ; i++ )  al->mantisse[i] = 0 ;
         al->signe = 0 ;
         return 0 ;
        }
 for( i = 0 ; i < n ; i++ )
        {
	 tempon = al->mantisse[i] ;
	 al->mantisse[i] = mul_add_div_ul(&carry , m , tempon , carry , BASE32 ) ;
        }

 if ( carry != 0 ) return 1 ; else return 0 ;
}

double depth_setMandelbrot(double xInf, double xSup, double yInf, double ySup )
{
 double temp , temp2 ;

 temp = fabs(xSup-xInf) ;
 if (temp == 0.0) return INFTY_DEPTH ;
 temp = ceil(-log10(temp)*10.0)*0.1 ;
 temp2 = fabs(ySup-yInf) ;
 if (temp2 == 0.0) return INFTY_DEPTH ;
 temp2 = ceil(-log10(temp2)*10.0)*0.1 ;
 if (temp < temp2) return temp ; else return temp2 ;
}

double depth_deci( Li* bf )
{
 long i = bf->taille_mantisse - 2 ;

 if ( bf->signe == 0 ) return INFTY_DEPTH ;
 while (( i >= 0 ) && ( bf->mantisse[i] == 0 )) i-- ;
 return ceil(10.0*(9.0*( bf->taille_mantisse - 2 - i )-log10(bf->mantisse[i]) ) )*0.1 ;
}

double depth_setMandelbrot_deci(char* xInf , char* xSup , char* yInf ,char* ySup )
{
 size_t longueur, longueur2 ;
 Li min, max , d ;
 int lecture_reel ;
 double profondeurX , profondeurY ;

 longueur  = strlen( xInf ) ;
 longueur2 = strlen( xSup ) ;
 if ( longueur > longueur2 ) longueur2 = longueur ;
 longueur  = strlen( yInf ) ;
 if ( longueur > longueur2 ) longueur2 = longueur ;
 longueur  = strlen( ySup ) ;
 if ( longueur > longueur2 ) longueur2 = longueur ;

 longueur2 = ( longueur2 / 9 ) + 6 ;

 //printf(" longueur2 = %i \n" , longueur2 ) ;

 lecture_reel  = create_BigFloat_deci2( &min , longueur2 , xInf ) ;
 lecture_reel += create_BigFloat_deci2( &max ,longueur2 , xSup ) ;

 //printLi( &min ) ;
 //printLi( &max ) ;
 //printf(" lecture_reel = %i \n" , lecture_reel ) ;

 if ( lecture_reel != 0 ) return INFTY_DEPTH ;
 createLi( &d , longueur2 ) ;
 sub_deci( &d , &max , &min ) ;

 profondeurX = depth_deci( &d ) ;

 destroyLi(&min) ;
 destroyLi(&max) ;
 lecture_reel += create_BigFloat_deci2( &min , longueur2 , yInf ) ;
 lecture_reel += create_BigFloat_deci2( &max , longueur2 , ySup ) ;
 if ( lecture_reel != 0 ) return INFTY_DEPTH ;
 sub_deci( &d , &min , &max ) ;
 profondeurY = depth_deci( &d ) ;
 destroyLi(&min) ;
 destroyLi(&max) ;
 destroyLi(&d);
 if ( profondeurX < profondeurY ) return profondeurX ;
                                  else
                                  return profondeurY ;
}

void add_2P_deci( Li* result , Li* a , Li* b )
// Je suppose que result est dj cre et a la taille de a et b.
// a et b sont deux nombres positifs de mme taille.
// Je ne gre pas le dbordement.
// result  <-  a + b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
                result->mantisse[i] = a->mantisse[i] ;
	 result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
                result->mantisse[i] = b->mantisse[i] ;
	 result->signe = b->signe ;
         return ;
        }

 result->signe = a->signe ;
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
         if ( en >= BASE32 )
                {
                 carry = 1L ;
                 result->mantisse[i] = en - BASE32 ;
                }
                else
                {
                 carry = 0L ;
                 result->mantisse[i] = en ;
                }
        }
}

void sub_2P_deci( Li* result , Li* a , Li* b )
// Je suppose que a et b sont positifs de mme taille.
// result  <-  a - b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
                result->mantisse[i] = a->mantisse[i] ;
         result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
                result->mantisse[i] = b->mantisse[i] ;
         result->signe = -b->signe ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         if ( en < 0 )
                {
                 carry = -1L ;
                 result->mantisse[i] = en + BASE32 ;
                }
                else
                {
                 carry = 0 ;
                 result->mantisse[i] = en ;
                }
        }
 if ( carry == -1L )
        {
         en = 0 ;
         while( result->mantisse[en] == 0 ) en++ ;
         result->mantisse[en] = BASE32 - result->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
         	result->mantisse[i] = CACHE32 - result->mantisse[i] ;
         result->signe = -(a->signe) ;
         return ;
        }

 result->signe = a->signe ;
 for( i = 0 ; i <= n - 1 ; i++ )
        if ( result->mantisse[i] != 0 ) break ;
 if ( i == n ) result->signe = 0 ;
}

void add_deci( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         add_2P_deci( result , a , b ) ;
         return ;
        }
 sub_2P_deci( result , a , b ) ;
}

void sub_deci( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_2P_deci( result , a , b ) ;
         return ;
        }
 add_2P_deci( result , a , b ) ;
}

void add_R_2P_deci( Li* a , Li* b )
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 ) return ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
	 	a->mantisse[i] = b->mantisse[i] ;
         a->signe = b->signe ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
	 if ( en >= BASE32 )
                {
                 carry = 1L ;
                 a->mantisse[i] = en - BASE32 ;
                }
                else
                {
                 carry = 0L ;
                 a->mantisse[i] = en ;
                }
        }
}

void sub_R_2P_deci( Li* a , Li* b )
//calcul a <-  a - b
// 2P signifie que cette routine travaille sur les valeurs absolues.
{
 long i , en , carry , n ;

 if ( b->signe == 0 ) return ;
 n = a->taille_mantisse ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) a->mantisse[i] = b->mantisse[i] ;
         a->signe = -b->signe ;
         return ;
        }

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         if ( en < 0 )
                {
                 carry = -1L ;
                 a->mantisse[i] = en + BASE32 ;
                }
                else
                {
                 carry = 0 ;
                 a->mantisse[i] = en ;
                }
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( a->mantisse[en]  == 0 ) en++ ;
         a->mantisse[en] = BASE32 - a->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                a->mantisse[i] = CACHE32 - a->mantisse[i] ;
         a->signe = -(a->signe) ;
         return ;
        }

 for( i = 0 ; i < n ; i++ )
        if ( a->mantisse[i] != 0 ) break ;
 if ( i == n ) a->signe = 0 ;
}

void square_2P_trunc_deci( Li* prod , Li* a )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
{
 long i , j , n , n_prod ;
 unsigned long p0 , p1 , p2 , p , tempon , tempon2 , aux , n2 , aux2 ;

 n = a->taille_mantisse ;
 n2 = (n << 1)-3 ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;

 for( i = n-2 ; i < n ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon2,p0, BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon,p0, BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon2,p0,BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	 p0 += prod->mantisse[i] ;
	 aux2 = p0 / BASE32 ;
	 prod->mantisse[i] = p0 - aux2 * BASE32 ;
	 p1 += (prod->mantisse[i+1] + aux2 ) ;
	 aux2 = p1 / BASE32 ;
	 prod->mantisse[i+1] = p1 - aux2 * BASE32 ;
	 p2 += ( prod->mantisse[i+2] + aux2 ) ;
	 prod->mantisse[i+2] = p2 % BASE32 ;
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	}



 for( i = n ; i < n2 ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n+1 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon2,p0,BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon,p0,BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_ul(&aux,tempon,tempon2,p0,BASE32 ) ;
		 p1 += aux ;
		 aux2 = p1 / BASE32 ;
		 p2 +=  aux2 ;
		 p1 -= ( aux2 * BASE32 ) ;
		}
	 p0 += prod->mantisse[i] ;
	 aux2 = p0 / BASE32 ;
	 prod->mantisse[i] = p0 - aux2 * BASE32 ;
	 p1 += (prod->mantisse[i+1] + aux2 ) ;
	 aux2 = p1 / BASE32 ;
	 prod->mantisse[i+1] = p1 - aux2 * BASE32 ;
	 p2 += ( prod->mantisse[i+2] + aux2 ) ;
	 prod->mantisse[i+2] = p2 % BASE32 ;
	}

 // i = 2n-3
 tempon  = a->mantisse[n-2] ;
 tempon2 = tempon << 1 ;
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_ul(&aux,tempon,tempon2,0,BASE32 ) ;
 p0 += prod->mantisse[n2] ;
 aux2 = p0 / BASE32 ;
 prod->mantisse[n2] = p0 -aux2 * BASE32 ;
 aux += (prod->mantisse[n2+1] + aux2 ) ;
 prod->mantisse[n2+1] = aux % BASE32 ;
 //  i = 2n-2
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_ul(&aux,tempon,tempon,0,BASE32 ) ;
 prod->mantisse[n2+1] = ( p0 + prod->mantisse[n2+1] ) % BASE32 ;
}


void mul_fft_Pollard_deci_trunc( Li* result , Li* a , Li* b , long k , unsigned long* prod ,
		    unsigned long* tp0, unsigned long* tp1, unsigned long* tp2 )
{
 long indiceMax , l , n ;

 if (( a->signe == 0) || ( b->signe == 0 ))
	{
	 result->signe = 0 ;
	 return ;
	}
 mul_fft_Pollard_deci(k,prod,(unsigned long*)a->mantisse,
                        (unsigned long*)b->mantisse,
	       tp0 , tp1 , tp2 ) ;
 result->signe = a->signe * b->signe ;
 n = a->taille_mantisse ;
 indiceMax = (n<<1)-2 ;
 for( l = n - 1 ; l <= indiceMax ; l++ )
	result->mantisse[l-n+1] = prod[l-1] ;
 for( l = 0 ; l < n ; l++ )
	if ( result->mantisse[l] != 0 ) break ;
 if ( l == n ) result->signe = 0 ;
}



void square_deci( Li* result , Li* prod , Li* a )
// Je suppose que a et b ont le mme nombre de chiffres ou nul et
// prod assez long pour contenir le rsultat du produit a*b.
{
 long indiceMax , l , n ;

 square_2P_trunc_deci( prod , a ) ;
 //square_2P_deci( prod , a ) ;
 result->signe = 1 ;
 n = a->taille_mantisse ;
 indiceMax = (n<<1)-2 ;
 for( l = n - 1 ; l <= indiceMax ; l++ )
	result->mantisse[l-n+1] = prod->mantisse[l-1] ; //*pa = *pp ;
 for( l = 0 ; l < n ; l++ )
	if ( result->mantisse[l] != 0 ) break ;
 if ( l == n ) result->signe = 0 ;
}

/*
void add_bin( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe == 1 )
        {
         add_2P_bin( result , a , b ) ;
         result->signe = a->signe ;
         return ;
        }
 if ( a->signe * b->signe == -1 )
        {
         sub_2P_bin( result , a , b ) ;
         result->signe *= a->signe ;
         return ;
        }
 add_2P_bin( result , a , b ) ;
 if ( a->signe == 0 ) result->signe = b->signe ;
        else result->signe = a->signe ;
}
*/
void add_R_deci( Li* a , Li* b )
// a <- a + b
{
 if ( a->signe * b->signe >= 0 )
        {
         add_R_2P_deci( a , b ) ;
         return ;
        }
 sub_R_2P_deci( a , b ) ;
}

void sub_R_deci( Li* a , Li* b )
// a <- a - b
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_R_2P_deci( a , b ) ;
         return ;
        }
 add_R_2P_deci( a , b ) ;
}

char* bigFloat_deciToStr( Li* li )
{
 char* S ;
 char* pS ;
 char pA[11] ;
 long taille_chaine = 2 ; // pour le . et le \0
 long en , i , k , j ;

 if ( li->signe == 0 )
	{
	 S = (char*) malloc(2 * sizeof(char) ) ;
	 *S = '0' ;
	 pS = S ;
	 pS++ ;
	 *pS = '\0' ;
	 return S ;
	}
 if ( li->signe == -1 ) taille_chaine++ ; // pour le signe -
 taille_chaine += ((li->taille_mantisse-1) * 9) ;
 S = (char*) malloc(taille_chaine * sizeof(char) ) ;
 pS = S ;
 if ( li->signe == -1 )
 	{
	 *pS = '-' ;
	 pS++ ;
	}
 en = li->mantisse[li->taille_mantisse-2] ;
 i = 0 ;
 do
 {
  pA[i] = '0' + (en % 10 ) ;
  en /= 10 ;
  i++ ;
 }
 while ( en != 0 ) ;
 for( k = i-1 ; k >= 0 ; k-- )
	{
	 *pS = pA[k] ;
	 pS++ ;
	}
 *pS = '.' ;
 pS++ ;
 for( j = li->taille_mantisse-3 ; j >= 0 ; j-- )
	{
	 en = li->mantisse[j] ;
	 for( i = 0 ; i < 9 ; i++ )
	 	{
		 pA[i] = '0' + (en % 10 ) ;
		 en /= 10 ;
		}
	 for( k = 8 ; k >= 0 ; k-- )
		{
		 *pS = pA[k] ;
		 pS++ ;
		}
	}

 *pS = '\0' ;
 return S ;
}

//--------------------------------------------------------------------
// -------   Entier long en base 2^30      ---------------------------
//--------------------------------------------------------------------

void add_R_2P_bin( Li* a , Li* b )
// a et b de mme taille et positif.
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 ) return ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ )
	 	a->mantisse[i] = b->mantisse[i] ;
         a->signe = b->signe ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
         carry = en >> DECALAGE32_BIN ;
         a->mantisse[i] = en & CACHE32_BIN ; // cache = 2^30 - 1
        }
}

void add_2P_bin( Li* result , Li* a , Li* b )
// Je suppose que result est dj cre et a la taille de a et b.
// a et b sont deux nombres positifs.
// Je ne gre pas le dbordement.
// result  <-  a + b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = a->mantisse[i] ;
	 result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = b->mantisse[i] ;
	 result->signe = b->signe ;
         return ;
        }

 result->signe = a->signe ;
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] + b->mantisse[i] + carry ;
         carry = en >> DECALAGE32_BIN ;
         result->mantisse[i] = en & CACHE32_BIN ;
	 //printf("%ld ** %ld %ld == %ld \n" , en , carry , result->Mantisse[i], CACHE32_BIN ) ;
        }
}

void sub_2P_bin( Li* result , Li* a , Li* b )
// Je suppose que a et b sont positifs.
// result  <-  a - b
{
 long i , en , carry , n ;

 n = a->taille_mantisse ;
 if ( b->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = a->mantisse[i] ;
         result->signe = a->signe ;
         return ;
        }
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = b->mantisse[i] ;
         result->signe = -(b->signe) ;
         return ;
        }
 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         result->mantisse[i] = en - carry * BASE32_BIN ;
	 //result->Mantisse[i] = en + (65536 >> ( 17 - (carry & 17) ) ) ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( result->mantisse[en] == 0 ) en++ ;
         result->mantisse[en] = BASE32_BIN - result->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                result->mantisse[i] = CACHE32_BIN - result->mantisse[i] ;
         result->signe = -(a->signe) ;
         return ;
        }

 result->signe = a->signe ;
 for( i = 0 ; i < n ; i++ )
        if ( result->mantisse[i] != 0 ) break ;
 if ( i == n ) result->signe = 0 ;
}

void sub_R_2P_bin( Li* a , Li* b )
// Je suppose que a et b sont positifs.
// a  <-  a - b
{
 long i , en , carry , n ;

 if ( b->signe == 0 ) return ;
 n = a->taille_mantisse ;
 if ( a->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) a->mantisse[i] = b->mantisse[i] ;
         a->signe = -(b->signe) ;
         return ;
        }

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a->mantisse[i] - b->mantisse[i] + carry ;
         carry = en >> 31 ;
         a->mantisse[i] = en - carry * BASE32_BIN ;
	 //a->Mantisse[i] = en + (65536 >> ( 17 - (carry & 17) ) ) ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( a->mantisse[en]  == 0 ) en++ ;
         a->mantisse[en] = BASE32_BIN - a->mantisse[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                a->mantisse[i] = CACHE32_BIN - a->mantisse[i] ;
         a->signe = -(a->signe) ;
         return ;
        }

 for( i = 0 ; i < n ; i++ )
        if ( a->mantisse[i] != 0 ) break ;
 if ( i == n ) a->signe = 0 ;
}

void add_bin( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         add_2P_bin( result , a , b ) ;
         return ;
        }
 sub_2P_bin( result , a , b ) ;
}

void add_R_bin( Li* a , Li* b )
// a <- a + b
{
 if ( a->signe * b->signe >= 0 )
        {
         add_R_2P_bin( a , b ) ;
         return ;
        }
 sub_R_2P_bin( a , b ) ;
}

void sub_bin( Li* result , Li* a , Li* b )
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_2P_bin( result , a , b ) ;
         return ;
        }
 add_2P_bin( result , a , b ) ;
}

void sub_R_bin( Li* a , Li* b )
// a <- a - b
{
 if ( a->signe * b->signe >= 0 )
        {
         sub_R_2P_bin( a , b ) ;
         return ;
        }
 add_R_2P_bin( a , b ) ;
}


void mulB_bin( Li* prod , Li* li , long m )
// multiplication par un chiffre.
// je ne gre pas le dbordement.
{
 long i , n ;
 unsigned long q  ;
 unsigned long aux ;

 n = li->taille_mantisse ;
 if (( li->signe == 0 ) || ( m == 0 ))
        {
         prod->signe = 0 ;
         for(i = 0 ; i < n ; i++ ) prod->mantisse[i]  = 0 ;
         return ;
        }
 prod->signe = li->signe ;
 q = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
	 aux = li->mantisse[i] ;
	 prod->mantisse[i] = mul_add_div_normalized_bin(&q,aux,m,q) ;
        }
}

long div_by_pow2_bin( Li* result , Li* li , long p )
// if faut que 0 < 2^p < BASE32_BIN.
// return the rest of the division.
{
 long i , a , r , q , n ;
 unsigned long tempon ;

 n = li->taille_mantisse ;
 if ( li->signe == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->mantisse[i] = 0 ;
         return 0 ;
        }
 result->signe = li->signe ;
 a = 1 ;
 a <<= p ;

 q = (li->mantisse[n - 1]) >> p ;
 r = (li->mantisse[n - 1]) & (a-1) ;
 result->mantisse[n - 1] = q ;

 for( i = n - 2 ; i >= 0 ; --i )
        {
	 tempon = li->mantisse[i] ;
	 r = mul_add_div(&q,r,BASE32_BIN,tempon,a);
         result->mantisse[i] = q ;
        }
 for( i = 0 ; i < n ; ++i )
        if ( result->mantisse[i]  != 0 ) break ;
 if ( i == n ) result->signe = 0 ;
 return r ;
}

/*
long div_by_pow2_Bin( Li* result , Li* li , long p )
// if faut que 0 < 2^p < BASE32_BIN.
// return the rest of the division.
{
 long i , a , r , q , cache , n ;

 n = li->Precision ;
 if ( li->Sign == 0 )
        {
         for( i = 0 ; i < n ; i++ ) result->Mantisse[i] = 0 ;
         return 0 ;
        }
 result->Sign = li->Sign ;
 cache = 1L ;
 cache <<= p ;
 cache -= 1L ;
 q = (li->Mantisse[n - 1]) >> p ;
 r = (li->Mantisse[n - 1]) & cache ;
 result->Mantisse[n - 1] = q ;

 for( i = n - 2 ; i >= 0 ; --i )
        {
	 a = li->Mantisse[i] + (r << DECALAGE32_BIN) ;   // DECALAGE32_BIN  16
         q = a >> p ;
         r = a & cache ;
         result->Mantisse[i] = q ;
        }
 for( i = 0 ; i < n ; ++i )
        if ( result->Mantisse[i]  != 0 ) break ;
 if ( i == n ) result->Sign = 0 ;
 return r ;
}
*/

void mul_2P_bin( Li* prod , Li* a , Li* b )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long m , carry=0 ,tempon , aux ;

 n = a->taille_mantisse ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         m = b->mantisse[i] ;
         carry = 0 ;
         for( j = 0 ; j < n ; j++ )
                {
		 tempon = carry + prod->mantisse[i+j] ;
		 aux = a->mantisse[j] ;
		 prod->mantisse[i+j] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
                }
         prod->mantisse[i+n] = carry ;
        }
}

void mul_2P_trunc_bin( Li* prod , Li* a , Li* b )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long m , carry=0 ,tempon , aux ;

 n = a->taille_mantisse ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;

 for( i = 0 ; i < n-1 ; i++ )
        {
         m = b->mantisse[i] ;
         carry = 0 ;
         for( j = n-2-i ; j < n ; j++ )
                {
		 tempon = carry + prod->mantisse[i+j] ;
		 aux = a->mantisse[j] ;
		 prod->mantisse[i+j] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
                }
         prod->mantisse[i+n] = carry ;
        }
 m = b->mantisse[n-1] ;
 carry = 0 ;
 for( j = 0 ; j < n ; j++ )
	{
	 tempon = carry + prod->mantisse[n-1+j] ;
	 aux = a->mantisse[j] ;
	 prod->mantisse[n-1+j] = mul_add_div_normalized_bin(&carry,m,aux,tempon) ;
	}
 prod->mantisse[n+n-1] = carry ;
}

void square_2P_bin( Li* prod , Li* a )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long p0 , p1 , p2 , p , tempon , tempon2 , aux , n2 ;

 n = a->taille_mantisse ;
 n2 = (n << 1)-3 ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;

 for( i = 0 ; i < n ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	}
 for( i = n ; i < n2 ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n+1 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	}

 // i = 2n-3
 tempon  = a->mantisse[n-2] ;
 tempon2 = tempon << 1 ;
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,0) ;
 p0 += prod->mantisse[n2] ;
 prod->mantisse[n2] = p0 & CACHE32_BIN ;
 aux += (prod->mantisse[n2+1] + ( p0 >> DECALAGE32_BIN ) ) ;
 prod->mantisse[n2+1] = aux & CACHE32_BIN ;
 //  i = 2n-2
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,0) ;
 prod->mantisse[n2+1] = ( p0 + prod->mantisse[n2+1] ) & CACHE32_BIN ;
}

void square_2P_trunc_bin( Li* prod , Li* a )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n , n_prod ;
 unsigned long p0 , p1 , p2 , p , tempon , tempon2 , aux , n2 ;

 n = a->taille_mantisse ;
 n2 = (n << 1)-3 ;
 n_prod = prod->taille_mantisse ;
 for( i = 0 ; i < n_prod ; i++ ) prod->mantisse[i] = 0 ;

 for( i = n-2 ; i < n ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	 //printf("p0 p1 p2 = %ld %ld % ld\n" , p0 , p1 , p2 ) ;
	}



 for( i = n ; i < n2 ; i++ )
	{
	 p = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n+1 ; j < p ; j++)
		{
		 tempon  = a->mantisse[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 if ( (i&1L) == 0 )  // i pair
	 	{
		 tempon  = a->mantisse[p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
		else  // i impair
		{
		 tempon  = a->mantisse[p] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a->mantisse[i-p] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod->mantisse[i] ;
	 prod->mantisse[i] = p0 & CACHE32_BIN ;
	 p1 += (prod->mantisse[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+1] = p1 & CACHE32_BIN ;
	 p2 += ( prod->mantisse[i+2] + ( p1 >> DECALAGE32_BIN ) ) ;
	 prod->mantisse[i+2] = p2 & CACHE32_BIN ;
	}

 // i = 2n-3
 tempon  = a->mantisse[n-2] ;
 tempon2 = tempon << 1 ;
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,0) ;
 p0 += prod->mantisse[n2] ;
 prod->mantisse[n2] = p0 & CACHE32_BIN ;
 aux += (prod->mantisse[n2+1] + ( p0 >> DECALAGE32_BIN ) ) ;
 prod->mantisse[n2+1] = aux & CACHE32_BIN ;
 //  i = 2n-2
 tempon  = a->mantisse[n-1] ;
 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,0) ;
 prod->mantisse[n2+1] = ( p0 + prod->mantisse[n2+1] ) & CACHE32_BIN ;
}

void square_bin( Li* result , Li* prod , Li* a )
// Je suppose que a et b ont le mme nombre de chiffres ou nul et
// prod assez long pour contenir le rsultat du produit a*b.
{
 long indiceMax , l , n ;

 square_2P_trunc_bin( prod , a ) ;
 //square_2P_Bin( prod , a ) ;
 result->signe = 1 ;
 n = a->taille_mantisse ;
 indiceMax = (n<<1)-2 ;
 for( l = n - 1 ; l <= indiceMax ; l++ )
	result->mantisse[l-n+1] = prod->mantisse[l] ; //*pa = *pp ;
 for( l = 0 ; l < n ; l++ )
	if ( result->mantisse[l] != 0 ) break ;
 if ( l == n ) result->signe = 0 ;
}

/*
void mul_Bin( Li* prod , Li* a , Li* b , long* indiceFin )
{
 mul_2P_Bin( prod , a , b , indiceFin ) ;
 prod->Sign = a->Sign * b->Sign ;
}
*/
void mul_bin( Li* result , Li* prod , Li* a , Li* b )
// Je suppose que a et b ont le mme nombre de chiffres ou nul et
// prod assez long pour contenir le rsultat du produit a*b.
{
 long indiceMax , l , n ;

 mul_2P_trunc_bin( prod , a , b ) ;
 //mul_2P_Bin( prod , a , b ) ;
 result->signe = a->signe * b->signe ;
 n = a->taille_mantisse ;
 indiceMax = (n<<1)-2 ;
 if ( result->signe != 0 )
 	{
         for( l = n - 1 ; l <= indiceMax ; l++ )
	 	result->mantisse[l-n+1] = prod->mantisse[l] ; //*pa = *pp ;
         for( l = 0 ; l < n ; l++ )
		if ( result->mantisse[l] != 0 ) break ;
	 if ( l == n ) result->signe = 0 ;
  	}
}

void mul_by2_R_bin( Li* a )
{
 long i , carry , en , n ;

 n = a->taille_mantisse ;
 carry = 0 ;
 for(i = 0 ; i < n ; i++  )
        {
         en    = ( (a->mantisse[i]) << 1 ) + carry ;
         carry = en >> DECALAGE32_BIN ;
         a->mantisse[i]  = en & CACHE32_BIN ;
        }
}

double BigFloatToFloat( Li* bfBaseBin )
{
 long i , n ;
 double r ;

 n = bfBaseBin->taille_mantisse ;
 if ( bfBaseBin->signe == 0 ) return 0.0 ;
 r = bfBaseBin->mantisse[0] / 1073741824.0 ;
 for(i = 1 ; i < n - 1 ; i++ )
        r = ( r + (bfBaseBin->mantisse[i])) / 1073741824.0 ;
 r += bfBaseBin->mantisse[n - 1 ] ;
 return (r * (bfBaseBin->signe) ) ;
}

void BigFloat_deciToBigFloat( Li* bf , Li* bfBaseBin )
{
 long i , n , n_Bin ;
 Li bfCopy ;

 n = bf->taille_mantisse ;
 n_Bin = bfBaseBin->taille_mantisse ;
 bfBaseBin->signe = bf->signe ;
 //copie de la partie entire.
 bfBaseBin->mantisse[n_Bin - 1] = bf->mantisse[n-2] ;
 // Maintenant on s'occupe de la partie fractionnaire.
 createLi( &bfCopy , bf->taille_mantisse ) ;
 copy( &bfCopy , bf ) ;
 bfCopy.mantisse[n - 2] = 0 ;
 for( i = n_Bin - 2 ; i >= 0 ; i-- )
        {
         mulB2_R_deci( &bfCopy , BASE32_BIN ) ;
         bfBaseBin->mantisse[i] = bfCopy.mantisse[n - 2] +
                                  BASE32 * bfCopy.mantisse[n - 1] ;
         bfCopy.mantisse[n - 2] = 0 ;
         bfCopy.mantisse[n - 1] = 0 ;
        }
 destroyLi( &bfCopy ) ;
}

//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------

void createLii( Lii* li , long nb )
{
 long i ;

 li->signe = 0 ;
 li->indiceMax = -1 ;
 li->taille_mantisse = nb ;
 li->mantisse = (long*) calloc( nb , sizeof(long) ) ;
 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;
}

void destroyLii( Lii* li )
{
 long i,nb=li->taille_mantisse ;

 for( i = 0 ; i < nb ; i++ ) li->mantisse[i] = 0 ;
 free( li->mantisse ) ;
 li->taille_mantisse = -10 ;
 li->signe = -1 ;
 li->indiceMax = -1 ;
}

void copyLii( Lii* nouveau , Lii* li )
// Je suppose que li et nouveau ont mme taille.
{
 long i , n ;

 nouveau->signe = li->signe ;
 nouveau->indiceMax = li->indiceMax ;
 n = li->taille_mantisse ;
 for( i = 0 ; i < n ; i++ ) nouveau->mantisse[i] = li->mantisse[i] ;
}

void printLii( Lii* li )
{
  long i , n ;

 n = li->taille_mantisse ;
 printf(" taille_mantisse = %ld\n" , n ) ;
 printf(" signe = %i\n" , li->signe ) ;
 printf(" indiceMax = %ld\n" , li->indiceMax ) ;
 for( i = n - 1 ; i >= 0 ; i-- ) printf("%ld " , li->mantisse[i] ) ;
 printf("\n") ;
}

//--------------------------------------------------------------------
// -------   Entier long en base 10^9  avec indiceMax     ------------
//--------------------------------------------------------------------

int initialise_fixedPoint_decii( Lii* a, long e )
// initialise le nombre dcimal a = e.0000...
// il faut que abs(e) < BASE32 where BASE32 = 10^9
{
 long n,signe ;

 n = a->taille_mantisse ;
 //for(i=0;i<n;i++) a->mantisse[i] = 0 ;
 if (n<=0) return 1 ; // The vector is not created
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return 0 ;
	}
 if (e<0)
	{
	 signe = -1 ;
	 e = -e ;
	}
	else signe = 1 ;

 a->signe = signe ;
 a->mantisse[n-1] = e  % BASE32 ;
 a->indiceMax = n-1 ;
 if (e/BASE32 > 0 )
	{
	 a->indiceMax = -1 ;
	 return 1 ;
	}
	else return 0 ;
}

int initialise_int_decii( Lii* a, long e )
// initialise le nombre dcimal a = e.0000...
// il faut que abs(e) < BASE32 where BASE32 = 10^9
{
 long n,signe ;

 n = a->taille_mantisse ;
 //for(i=0;i<n;i++) a->mantisse[i] = 0 ;
 if (n<=0) return 1 ; // The vector is not created
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return 0 ;
	}

 if (e<0)
	{
	 signe = -1 ;
	 e = -e ;
	}
	else signe = 1 ;

 a->signe = signe ;
 a->mantisse[0] = e  % BASE32 ;
 a->indiceMax = 0 ;
 e = e/BASE32 ;
 if (e==0) return 0 ;
 if (n>1)
	{
	 a->mantisse[1] = e ;
	 a->indiceMax = 1 ;
	 return 0 ;
	}
	else return 1 ;
}

long add_R_2P_decii_spe( Lii* a , Lii* b )
// a <- a + b
// I suppose that a > b >= 0 .
// return the overflow digit !!!
{
 long carry = 0 ;
 long i ;
 long aux ;
 long iMaxA = a->indiceMax ;
 long iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return 0 ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] + b->mantisse[i] + carry ;
	 if ( aux >= BASE32 )
                {
                 carry = 1L ;
                 a->mantisse[i] = aux - BASE32 ;
                }
                else
                {
                 carry = 0L ;
                 a->mantisse[i] = aux ;
                }
	}

 if ( carry == 0 ) return 0 ;
 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 if ( aux >= BASE32 )
                {
                 carry = 1L ;
                 a->mantisse[i] = aux - BASE32 ;
                }
                else
                {
                 carry = 0L ;
                 a->mantisse[i] = aux ;
                }
	 if ( carry == 0 ) break ;
	}
 if ( iMaxA < (a->taille_mantisse - 1))
	{
	 if ( carry == 1 )
		{
		 iMaxA++ ;
		 a->mantisse[iMaxA] += carry ;
		 a->indiceMax = iMaxA ;
		 return 0 ;
		}
		else return 0 ;
	}
	else return 1 ;
}

void sub_R_2P_decii_spe( Lii* a , Lii* b )
// a <- a - b
// Je suppose que a >= b >= 0 .
{
 long carry = 0 ;
 long i ;
 long aux ;
 long iMaxA = a->indiceMax ;
 long iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] - b->mantisse[i] + carry ;
	 if ( aux < 0 )
		{
		 a->mantisse[i] = aux + BASE32 ;
		 carry = -1L ;
		}
		else
		{
		 a->mantisse[i] = aux ;
		 carry = 0L ;
		}
	}

 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 if ( aux < 0 )
		{
		 a->mantisse[i] = aux + BASE32 ;
		 carry = -1L ;
		}
		else
		{
		 a->mantisse[i] = aux ;
		 carry = 0L ;
		}
	 if ( carry == 0 ) break ;
	}
 while (( iMaxA >=0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if ( iMaxA < 0 ) a->signe = 0 ;
}

long divB_R_decii( Lii* a , long D )
// Je suppose que 0 < D < BASE32.
{
 long i ;
 long iMaxA = a->indiceMax ;
 long r  ;
 unsigned long q , temp ;

 if ( a->signe == 0 ) return 0 ;
 r = 0 ;

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 r = mul_add_div_ul( &q,r,BASE32,temp,D ) ;
	 a->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( a->mantisse[0] == 0 )) a->signe = 0 ;
 	else a->signe = 1 ;

 return r ;
}

long divB_decii( Lii* res , Lii* a , long D )
// Je suppose que 0 < D < BASE32 .
// res <- a / D
{
 long i , r=0 ;
 long iMaxA = a->indiceMax ;
 unsigned long temp , q ;

 if ( a->signe == 0 )
	{
	 res->signe = 0 ;
	 return 0 ;
	}

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 r = mul_add_div_ul( &q,r,BASE32,temp,D ) ;
	 res->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( res->mantisse[iMaxA] == 0 )) iMaxA-- ;
 res->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( res->mantisse[0] == 0 ))
 	res->signe = 0 ; else res->signe = 1 ;

 return r ;
}

long mulB_R_decii( Lii* a , long m )
// Cette routine calcule le produit a <- m * a o
// 0 <= m < BASE32 .
{
 long i ;
 long iMaxA = a->indiceMax ;
 unsigned long temp, carry=0 ;

 if ( a->signe == 0 ) return 0 ;
 if ( m == 0 )
	{
	 a->signe = 0 ;
	 return 0 ;
	}

 for( i = 0 ; i <= iMaxA ; i++ )
	{
	 temp = a->mantisse[i] ;
	 a->mantisse[i] = mul_add_div_ul( &carry,m,temp,carry,BASE32 ) ;
	}

 if ( carry != 0 )
 	{
	 if ( (iMaxA + 1) == a->taille_mantisse ) return carry ;
		else
		{
		 a->indiceMax = iMaxA + 1 ;
		 a->mantisse[iMaxA+1] = carry ;
		 return 0 ;
		}
	}
	else return 0 ;
}

void print_cst_base10_decii( Lii* a )
// affiche e , pi , sqrt(2) ...
{
 long i = 1 ;
 long j , div , temp ;
 long nb_decimales = a->taille_mantisse - 1 ;

 printf("%ld,\n" , a->mantisse[nb_decimales] ) ;
 while( nb_decimales > 0 )
   {
     nb_decimales-- ;
     div = BASE32 / 10 ;
     temp = a->mantisse[nb_decimales] ;
     for( j = 1 ; j < 10 ; j++ )
       {
	 printf("%ld" , ( ( temp / div ) % 10 ) ) ;
	 div /= 10 ;
         if ( i % 100 == 0 ) printf("  :: %ld\n" , i ) ;
         if (( i % 10 == 0 ) && ( i % 100 != 0 )) printf(" ") ;
         i++ ;
       }
   }
 printf("\n\n") ;
}

//--------------------------------------------------------------------
// -------   Entier long en base 2^30  avec indiceMax     ------------
//--------------------------------------------------------------------

int initialise_fixedPoint_bini( Lii* a, long e )
// initialise le nombre dcimal a = e.0000...
// il faut que abs(e) < BASE32_BIN where BASE32_BIN = 2^30
{
 long n,signe ;

 n = a->taille_mantisse ;
 //for(i=0;i<n;i++) a->mantisse[i] = 0 ;
 if (n<=0) return 1 ; // The vector is not created
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return 0 ;
	}
 if (e<0)
	{
	 signe = -1 ;
	 e = -e ;
	}
	else signe = 1 ;

 a->signe = signe ;
 a->mantisse[n-1] = e  & CACHE32_BIN ;
 a->indiceMax = n-1 ;
 e = e >> DECALAGE32_BIN ;
 if (e > 0 )
	{
	 a->indiceMax = -1 ;
	 return 1 ;
	}
	else return 0 ;
}

int initialise_int_bini( Lii* a, long e )
{
 long n,signe ;

 n = a->taille_mantisse ;
 //for(i=0;i<n;i++) a->mantisse[i] = 0 ;
 if (n<=0) return 1 ; // The vector is not created
 if( e == 0 )
	{
	 a->signe = 0 ;
	 a->indiceMax = -1 ;
	 return 0 ;
	}

 if (e<0)
	{
	 signe = -1 ;
	 e = -e ;
	}
	else signe = 1 ;

 a->signe = signe ;
 a->mantisse[0] = e  & CACHE32_BIN ;
 a->indiceMax = 0 ;
 e >>= DECALAGE32_BIN ;
 if (e==0) return 0 ;
 if (n>1)
	{
	 a->mantisse[1] = e ;
	 a->indiceMax = 1 ;
	 return 0 ;
	}
	else return 1 ;
}

long add_R_2P_bini_spe( Lii* a , Lii* b )
// a <- a + b
// I suppose that a > b >= 0 .
// return the overflow digit !!!
{
 long carry = 0 ;
 long i ;
 long aux ;
 long iMaxA = a->indiceMax ;
 long iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return 0 ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] + b->mantisse[i] + carry ;
	 carry = aux >> DECALAGE32_BIN ;
	 a->mantisse[i] = aux & CACHE32_BIN ;
	}
 if ( carry == 0 ) return 0 ;
 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 carry = aux >> DECALAGE32_BIN ;
	 a->mantisse[i] = aux & CACHE32_BIN ;
	 if ( carry == 0 ) break ;
	}
 if ( iMaxA < (a->taille_mantisse - 1))
	{
	 if ( carry == 1 )
		{
		 iMaxA++ ;
		 a->mantisse[iMaxA] += carry ;
		 a->indiceMax = iMaxA ;
		 return 0 ;
		}
		else return 0 ;
	}
	else return 1 ;
}

void sub_R_2P_bini_spe( Lii* a , Lii* b )
// a <- a - b
// Je suppose que a >= b >= 0 .
{
 long carry = 0 ;
 long i ;
 long aux ;
 long iMaxA = a->indiceMax ;
 long iMaxB = b->indiceMax ;

 if ( b->signe ==  0 ) return ;
 for( i = 0 ; i <= iMaxB ; i++ )
	{
	 aux = a->mantisse[i] - b->mantisse[i] + carry ;
	 carry = aux >> 31 ;
	 a->mantisse[i] = aux - carry * BASE32_BIN ;
	}

 for( i = iMaxB+1 ; i <= iMaxA ; i++ )
	{
	 aux = a->mantisse[i] + carry ;
	 carry = aux >> 31 ;
	 a->mantisse[i] = aux - carry * BASE32_BIN ;
	 if ( carry == 0 ) break ;
	}
 while (( iMaxA >=0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if ( iMaxA < 0 ) a->signe = 0 ;
}

long divB_R_bini( Lii* a , long D )
// Je suppose que 0 < D < BASE32_BIN .
{
 long i ;
 long iMaxA = a->indiceMax ;
 long r  ;
 unsigned long q , temp ;

 if ( a->signe == 0 ) return 0 ;
 r = 0 ;

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 r = mulB_add_div_ul( &q,r,temp,D ) ;
	 //r = mul_add_div_ul( &q,r,BASE32_BIN,temp,D ) ;
	 a->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( a->mantisse[iMaxA] == 0 )) iMaxA-- ;
 a->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( a->mantisse[0] == 0 )) a->signe = 0 ;
 	else a->signe = 1 ;

 return r ;
}

long divB_bini( Lii* res , Lii* a , long D )
// Je suppose que 0 < D < BASE32_BIN .
// res <- a / D
{
 long i , r=0 ;
 long iMaxA = a->indiceMax ;
 unsigned long temp , q ;

 if ( a->signe == 0 )
	{
	 res->signe = 0 ;
	 return 0 ;
	}

 for( i = iMaxA ; i >= 0 ; i-- )
	{
	 temp = a->mantisse[i] ;
	 r = mulB_add_div_ul( &q,r,temp,D ) ;
	 //r = mul_add_div_ul( &q,r,BASE32_BIN,temp,D ) ;
	 res->mantisse[i] = q ;
	}

 while (( iMaxA >= 0 ) && ( res->mantisse[iMaxA] == 0 )) iMaxA-- ;
 res->indiceMax = iMaxA ;
 if (( iMaxA <= 0 ) && ( res->mantisse[0] == 0 ))
 	res->signe = 0 ; else res->signe = 1 ;

 return r ;
}

long mulB_R_bini( Lii* a , long m )
// Cette routine calcule le produit a <- m * a o
// 0 <= m < BASE32_BIN .
{
 long i ;
 long iMaxA = a->indiceMax ;
 unsigned long temp, carry=0 ;

 if ( a->signe == 0 ) return 0 ;
 if ( m == 0 )
	{
	 a->signe = 0 ;
	 return 0 ;
	}
 for( i = 0 ; i <= iMaxA ; i++ )
	{
	 temp = a->mantisse[i] ;
	 a->mantisse[i] = mul_add_div_normalized_bin( &carry,m,temp,carry ) ;
	}


 if ( carry != 0 )
 	{
	 if ( (iMaxA + 1) == a->taille_mantisse ) return carry ;
		else
		{
		 a->indiceMax = iMaxA + 1 ;
		 a->mantisse[iMaxA+1] = carry ;
		 return 0 ;
		}
	}
	else return 0 ;
}

void print_cst_base2_bini( Lii* a )
// affiche e , pi , sqrt(2) ... en base 2.
{
 long i = 1 ;
 long j , temp , exposant , temp2 ;
 long n = a->taille_mantisse - 1 ;


 exposant  = 0 ;
 temp  = a->mantisse[n] ;
 temp2 = temp ;
 for( exposant = 0 ; temp2 != 0 ; exposant++ ) temp2 >>= 1 ;
 exposant-- ;
 while( exposant >= 0 )
   {
     printf("%ld" , ((temp >> exposant) & PARITE)  ) ;
     exposant-- ;
   }

 printf( ",\n" ) ;
 j = n - 1 ;
 i = 1 ;
 while( j >= 0 )
   {
     temp  = a->mantisse[j] ;
     for( exposant = DECALAGE32_BIN - 1 ; exposant >= 0 ; exposant-- )
       {
	 printf("%ld" , ((temp >> exposant) & PARITE ) ) ;
	 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
         if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
	 i++ ;
       }
     j-- ;
   }
 printf("\n\n") ;
}

void print_cst_base16_bini( Lii* a )
// affiche e , pi , sqrt(2) ... en base 16.
{
 long i = 1 ;
 long j , exposant , temp , temp2 , pariteDeb ;
 long n = a->taille_mantisse - 1 ;
 int chiffre ;

  exposant  = 0 ;
 temp  = a->mantisse[n] ;
 temp2 = temp ;
 for( exposant = 0 ; temp2 != 0 ; exposant++ ) temp2 >>= 4 ;
 exposant-- ;
 while( exposant >= 0 )
   {
     printf("%ld" , ((temp >> ( exposant << 4 )) & CACHE_BASE16 )  ) ;
     exposant-- ;
   }

 printf( ",\n" ) ;
 j = n - 1 ;
 pariteDeb = j & PARITE ;
 i = 1 ;
 while( j >= 0 )
   {
    temp  = a->mantisse[j] ;
    if ( (j & PARITE) == pariteDeb )
    	{
    	 for( exposant = DECALAGE32_BIN - 4 ; exposant >= 0 ; exposant -= 4 )
         	{
		 chiffre = (temp >> exposant )  & CACHE_BASE16 ;
		 //printf("%i\n" , chiffre ) ;
		 if ( chiffre < 10 ) printf("%i" , chiffre ) ;
		 	        else printf("%c" , chiffre + 55 ) ;
		 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
        	 if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
		 i++ ;
       		}
	 temp2 = temp & 3L ;
	}
	else
	{
	 chiffre = (temp2 << 2) + (temp >> (DECALAGE32_BIN - 2) ) ;
	 //printf("%i\n" , chiffre ) ;
	 if ( chiffre < 10 ) printf("%i" , chiffre ) ;
	 	        else printf("%c" , chiffre + 55 ) ;
	 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
  	 if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
	 i++ ;
    	 for( exposant = DECALAGE32_BIN - 6 ; exposant >= 0 ; exposant -= 4 )
         	{
		 chiffre = (temp >> exposant )  & CACHE_BASE16 ;
		 //printf("%i\n" , chiffre ) ;
		 if ( chiffre < 10 ) printf("%i" , chiffre ) ;
		 	else printf("%c" , chiffre + 55 ) ;
		 if ( (i % 100) == 0 ) printf("  :: %ld \n" , i ) ;
        	 if (( (i % 10) == 0 ) && ( (i % 100) != 0 )) printf(" ") ;
		 i++ ;
       		}
	}
     j-- ;
   }
 printf("\n\n") ;
}



//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^30  de bas niveau     ----------------
//-----------------------------------------------------------------------------

void mul_2P_naif_bin( long* prod , long* a , long* b , long n ) // suppose que prod est dj cr
{
 long i , j ,n_prod;
 unsigned long m , carry=0 ,tempon , tempon2 ;

 n_prod = n << 1 ;
 for( i = 0 ; i < n_prod ; i++ ) prod[i] = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         m = b[i] ;
         carry = 0 ;
         for( j = 0 ; j < n ; j++ )
                {
		 tempon  = a[j] ;
		 tempon2 = carry + prod[i+j] ;
		 prod[i+j] = mul_add_div_normalized_bin( &carry,m,tempon,tempon2 ) ;
                }
         prod[i+n] += carry ;
        }
}

void square_2P_naif_bin( long* prod , long* a , long n )
// Je suppose que prod est assez grand pour contenir le produit a * b .
// si a = 0 ou b = 0 je fais quand mme le produit. Le traitement du signe
// se fera dans la routine du cas gnral.
// Je suppose que a et b ont le mme nombre de chiffres.
{
 long i , j , n1 , n2 , ih , i1 ;
 unsigned long p0 , p1 , p2 , tempon , tempon2 , aux ;

 n1 = n - 1 ;
 n2 = (n << 1) ;
 for( i = 0 ; i < n2 ; i++ ) prod[i] = 0 ;
 n2 -= 2 ;
 for( i = 0 ; i < n1 ; i+=2 )
	{
	 ih = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j < ih ; j++)
		{
		 tempon  = a[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 tempon  = a[ih] ;
	 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
	 p1 += aux ;
	 p2 += (p1 >> DECALAGE32_BIN ) ;
	 p1 &= CACHE32_BIN ;
	 p0 += prod[i] ;
	 prod[i] = p0 & CACHE32_BIN ;
	 p1 += (prod[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod[i+1] = p1 & CACHE32_BIN ;
	 prod[i+2] += p2 + ( p1 >> DECALAGE32_BIN )  ;

	 i1 = i + 1 ;
	 p0=p1=p2=0 ;
	 for( j = 0 ; j <= ih ; j++)
		{
		 tempon  = a[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a[i1-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod[i1] ;
	 prod[i1] = p0 & CACHE32_BIN ;
	 p1 += (prod[i1+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod[i1+1] = p1 & CACHE32_BIN ;
	 prod[i1+2] += p2 + ( p1 >> DECALAGE32_BIN ) ;
	}

 for( ; i < n2 ; i+=2 )
	{
	 ih = i >> 1 ;
	 p0=p1=p2=0 ;
	 for( j = i-n1 ; j < ih ; j++)
		{
		 tempon  = a[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a[i-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 tempon  = a[ih] ;
	 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
	 p1 += aux ;
	 p2 += (p1 >> DECALAGE32_BIN ) ;
	 p1 &= CACHE32_BIN ;
	 p0 += prod[i] ;
	 prod[i] = p0 & CACHE32_BIN ;
	 p1 += (prod[i+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod[i+1] = p1 & CACHE32_BIN ;
	 prod[i+2] += p2 + ( p1 >> DECALAGE32_BIN ) ;

	 i1 = i + 1 ;
	 p0=p1=p2=0 ;
	 for( j = i1-n1 ; j <= ih ; j++)
		{
		 tempon  = a[j] ;
		 tempon2 = tempon << 1 ;
		 tempon  = a[i1-j] ;
		 p0 = mul_add_div_normalized_bin(&aux,tempon,tempon2,p0) ;
		 p1 += aux ;
		 p2 += (p1 >> DECALAGE32_BIN ) ;
		 p1 &= CACHE32_BIN ;
		}
	 p0 += prod[i1] ;
	 prod[i1] = p0 & CACHE32_BIN ;
	 p1 += (prod[i1+1] + ( p0 >> DECALAGE32_BIN ) ) ;
	 prod[i1+1] = p1 & CACHE32_BIN ;
	 prod[i1+2] += p2 + ( p1 >> DECALAGE32_BIN )  ;
	}

 //  i = 2n-2
 tempon  = a[n1] ;
 p0 = prod[n2] ;
 prod[n2] = mul_add_div_normalized_bin(&aux,tempon,tempon,p0) ;
 prod[n2+1] += aux ;
}












long** createMatrix( long n )
{
 long i ;
 long ** t ;

 t = ( long **) malloc( n * sizeof( long* ) ) ;
 if ( t == NULL ) return NULL ;

 for( i = 0 ; i < n ; i++ )
        {
         *( t + i ) = (long*) calloc( n , sizeof( long ) ) ;
         if ( *( t + i ) == NULL )
                {
                 while ( --i >= 0  )  free( *( t + i ) ) ;
                 return NULL ;
                }
        }
 return t ;
}

void destroyMatrix( long** ma , long n )
{
 if ( ma != NULL )
        {
         while( --n >= 0 ) free( *( ma + n ) ) ;
         free( ma ) ;
        }
}

// je suppose que BinRe, BinIm, BinCx et BinCy ont la mme taille.
long calcul_germe_mandelbrotNorm1(Li* BinRe , Li* BinIm , Li* BinCx , Li* BinCy , long iter ,
			     	  Li* a0 , Li* a1 , Li* a2 , Li* prod , long tailleBin )
{
 long k , n ;

 n = tailleBin-1 ;

 for( k = 0 ; k < iter ; k++ )
 	{
	 sub_bin( a0 , BinRe , BinIm ) ;
 	 add_bin( a1 , BinRe , BinIm ) ;
   	 if ( BinIm->signe * BinRe->signe >= 0 )
     		{
       		 if ( a1->mantisse[n] >= 3  ) break ;
          	}
           	else
            	{
             	 if ( a0->mantisse[n] >= 3 ) break ;
               	}

	 mul_bin( a2 , prod , a0 , a1 ) ;
  	 add_R_bin( a2 , BinCx ) ;
	 mul_bin( BinIm , prod , BinRe , BinIm ) ;
  	 mul_by2_R_bin( BinIm ) ;
    	 copy( BinRe , a2 ) ;
      	 add_R_bin( BinIm , BinCy ) ;
        }
 return k ;
}

// je suppose que BinRe, BinIm, BinCx et BinCy ont la mme taille.
long calcul_germe_mandelbrot(Li* BinRe , Li* BinIm , Li* BinCx , Li* BinCy , long iter ,
			     Li* a0 , Li* a1 , Li* a2 , Li* prod , long tailleBin )
{
 long kk , nn , indiceMax ;

 nn = tailleBin-1 ;
 indiceMax = (tailleBin<<1)-2 ;

 for( kk = 0 ; kk < iter ; kk++ )
 	{
	 //mul_Bin( a0 , prod , BinRe , BinRe ) ;
	 //mul_Bin( a1 , prod , BinIm , BinIm ) ;
	 square_bin( a0 , prod , BinRe ) ;
	 square_bin( a1 , prod , BinIm ) ;

	 add_bin( a2 , a0 , a1 ) ;
	 if (a2->mantisse[nn] >= 4) break ;
	 sub_R_bin( a0 , a1 ) ;
	 add_R_bin( a0 , BinCx ) ;
	 mul_bin( BinIm , prod , BinIm , BinRe ) ;
	 mul_by2_R_bin( BinIm ) ;
	 copy( BinRe , a0 ) ;
	/*
	 BinRe->signe = a0->signe ;
	 for( i = 0 ; i < tailleBin ; i++ )
	 	BinRe->mantisse[i] = a0->mantisse[i] ;
	*/
	 add_R_bin( BinIm , BinCy ) ;
	/*
	 if ( BinIm->signe * BinCy->signe == 1 )
		{
		 add_R_2P_Bin( BinIm , BinCy ) ;
		}
		else
		{
		 if ( BinIm->signe * BinCy->signe == -1 )
			{
			 s = BinIm->signe ;
			 sub_R_2P_Bin( BinIm , BinCy ) ;
			 BinIm->signe *= s ;
			}
			else
			 add_R_2P_Bin( BinIm , BinCy ) ;
        	}
	*/
	}
 return kk ;
}

int calcul_MandelbrotNorm1( char* xInf ,
                              char* xSup ,
                              char* yInf ,
                              char* ySup ,
			      long tailleBin,
                              long iter ,
                              long **hypert )
{
 Li Min ;
 Li Max ;
 Li d   ;
 long i, j , taille ;
 Li  BinRe  ;
 Li  BinIm  ;
 Li  BinCx  ;
 Li  BinCy  ;
 Li  BinDx  ;
 Li  BinDy  ;
 Li  BinAux ;
//Li_9  BinX  ;
 Li  BinY   ;
 Li BinYMax ;
 Li BinXMin ;
 Li a0 ;
 Li a1 ;
 Li a2 ;
 Li prod ;
 //clock_t tops ,tops2 ;


 taille = (long int) ceil( 1.00343331888 * tailleBin ) + 8 ;
 create_BigFloat_deci2( &Min , taille , xInf ) ;
 create_BigFloat_deci2( &Max , taille , xSup ) ;
 createLi( &d , taille ) ;
 sub_deci( &d , &Max , &Min ) ;
 createLi( &BinDx , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDx ) ;
 createLi( &BinAux , tailleBin ) ;
 div_by_pow2_bin( &BinAux , &BinDx , 9 ) ;
 copy( &BinDx , &BinAux ) ;
 createLi( &BinXMin , tailleBin ) ;
 BigFloat_deciToBigFloat( &Min , &BinXMin ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 create_BigFloat_deci2( &Min , taille , yInf ) ;
 create_BigFloat_deci2( &Max , taille , ySup ) ;
 sub_deci( &d , &Min , &Max ) ;
 createLi( &BinDy , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDy ) ;
 div_by_pow2_bin( &BinAux , &BinDy , 9 ) ;
 copy( &BinDy , &BinAux ) ;
 createLi( &BinYMax , tailleBin ) ;
 BigFloat_deciToBigFloat( &Max , &BinYMax ) ;

 createLi( &BinY , tailleBin ) ;
 createLi( &BinCx , tailleBin ) ;
 createLi( &BinCy , tailleBin ) ;
 createLi( &BinRe , tailleBin ) ;
 createLi( &BinIm , tailleBin ) ;
 createLi( &a0 , tailleBin ) ;
 createLi( &a1 , tailleBin ) ;
 createLi( &a2 , tailleBin ) ;
 createLi( &prod , 2*tailleBin+2 ) ; // pas de souci ;)

 //tops = clock() ;

 for( i = 0 ; i < 512 ; i++ )
        {
	 mulB_bin( &BinAux  , &BinDy , i ) ;
         add_bin( &BinY , &BinYMax , &BinAux ) ;
         for( j = 0 ; j < 512 ; j++ )
                {
		 mulB_bin( &BinAux , &BinDx , j ) ;
                 add_bin( &BinRe , &BinXMin , &BinAux ) ;
		 copy( &BinIm , &BinY ) ;
                 copy( &BinCx , &BinRe ) ;
                 copy( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrotNorm1( &BinRe , &BinIm , &BinCx , &BinCy , iter ,
		 						    &a0 , &a1 , &a2 , &prod , tailleBin) ;
                }
        }

 //tops2 = clock() ;
 //printf(" Dure du calcul = %10.3f secondes.\n" , (double) (tops2-tops)/ CLOCKS_PER_SEC ) ;

 destroyLi( &a0) ;
 destroyLi( &a1) ;
 destroyLi( &a2) ;
 destroyLi( &prod) ;
 destroyLi( &BinRe ) ;
 destroyLi( &BinIm ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 destroyLi( &BinCx ) ;
 destroyLi( &BinCy ) ;
 destroyLi( &BinYMax ) ;
 destroyLi( &BinXMin ) ;
 destroyLi( &BinY ) ;
 destroyLi( &BinDx) ;
 destroyLi( &BinDy ) ;
 destroyLi( &BinAux ) ;

 return 0 ;
}

int calcul_Mandelbrot( char* xInf ,char* xSup ,char* yInf ,char* ySup ,
		       long tailleBin,long iter ,long **hypert )
{
 Li Min ;
 Li Max ;
 Li d   ;
 long i, j , taille ;
 Li  BinRe  ;
 Li  BinIm  ;
 Li  BinCx  ;
 Li  BinCy  ;
 Li  BinDx  ;
 Li  BinDy  ;
 Li  BinAux ;
 Li  BinY   ;
 Li BinYMax ;
 Li BinXMin ;
 Li a0 ;
 Li a1 ;
 Li a2 ;
 Li prod ;
 //clock_t tops ,tops2 ;


 taille = (long int) ceil( 1.00343331888 * tailleBin ) + 8 ;
 create_BigFloat_deci2( &Min , taille , xInf ) ;
 create_BigFloat_deci2( &Max , taille , xSup ) ;
 createLi( &d , taille ) ;
 sub_deci( &d , &Max , &Min ) ;
 createLi( &BinDx , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDx ) ;
 createLi( &BinAux , tailleBin ) ;
 div_by_pow2_bin( &BinAux , &BinDx , 9 ) ;
 copy( &BinDx , &BinAux ) ;
 createLi( &BinXMin , tailleBin ) ;
 BigFloat_deciToBigFloat( &Min , &BinXMin ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 create_BigFloat_deci2( &Min , taille , yInf ) ;
 create_BigFloat_deci2( &Max , taille , ySup ) ;
 sub_deci( &d , &Min , &Max ) ;
 createLi( &BinDy , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDy ) ;
 div_by_pow2_bin( &BinAux , &BinDy , 9 ) ;
 copy( &BinDy , &BinAux ) ;
 createLi( &BinYMax , tailleBin ) ;
 BigFloat_deciToBigFloat( &Max , &BinYMax ) ;

 createLi( &BinY , tailleBin ) ;
 createLi( &BinCx , tailleBin ) ;
 createLi( &BinCy , tailleBin ) ;
 createLi( &BinRe , tailleBin ) ;
 createLi( &BinIm , tailleBin ) ;
 createLi( &a0 , tailleBin ) ;
 createLi( &a1 , tailleBin ) ;
 createLi( &a2 , tailleBin ) ;
 createLi( &prod , 2*tailleBin+2 ) ; // pas de souci ;)

 //tops = clock() ;

 for( i = 0 ; i < 512 ; i++ )
        {
	 mulB_bin( &BinAux  , &BinDy , i ) ;
         add_bin( &BinY , &BinYMax , &BinAux ) ;
         for( j = 0 ; j < 512 ; j++ )
                {
		 mulB_bin( &BinAux , &BinDx , j ) ;
                 add_bin( &BinRe , &BinXMin , &BinAux ) ;
		 copy( &BinIm , &BinY ) ;
                 copy( &BinCx , &BinRe ) ;
                 copy( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ,
		 						    &a0 , &a1 , &a2 , &prod , tailleBin) ;

                }
        }

 //tops2 = clock() ;
 //printf(" Dure du calcul = %10.3f secondes.\n" , (double) (tops2-tops)/ CLOCKS_PER_SEC ) ;

 destroyLi( &a0) ;
 destroyLi( &a1) ;
 destroyLi( &a2) ;
 destroyLi( &prod) ;
 destroyLi( &BinRe ) ;
 destroyLi( &BinIm ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 destroyLi( &BinCx ) ;
 destroyLi( &BinCy ) ;
 destroyLi( &BinYMax ) ;
 destroyLi( &BinXMin ) ;
 destroyLi( &BinY ) ;
 destroyLi( &BinDx) ;
 destroyLi( &BinDy ) ;
 destroyLi( &BinAux ) ;

 return 0 ;
}

// Je suppose que sizeMatrix, la taille de la matrice, est une puissance de 2.
int calcul_Mandelbrot2( char* xInf ,char* xSup ,char* yInf ,char* ySup ,
		           long tailleBin,long iter ,long **hypert, long sizeMatrix )
{
 Li Min ;
 Li Max ;
 Li d   ;
 long i, j , taille , powerMatrix ;
 Li  BinRe  ;
 Li  BinIm  ;
 Li  BinCx  ;
 Li  BinCy  ;
 Li  BinDx  ;
 Li  BinDy  ;
 Li  BinAux ;
 Li  BinY   ;
 Li BinYMax ;
 Li BinXMin ;
 Li a0 ;
 Li a1 ;
 Li a2 ;
 Li prod ;


 powerMatrix = 0 ;
 i = sizeMatrix ;
 while( i > 1 )
	{
	 i >>= 1 ;
	 powerMatrix++ ;
	}

 taille = (long int) ceil( 1.00343331888 * tailleBin ) + 8 ;
 create_BigFloat_deci2( &Min , taille , xInf ) ;
 create_BigFloat_deci2( &Max , taille , xSup ) ;
 createLi( &d , taille ) ;
 sub_deci( &d , &Max , &Min ) ;
 createLi( &BinDx , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDx ) ;
 createLi( &BinAux , tailleBin ) ;
 div_by_pow2_bin( &BinAux , &BinDx , powerMatrix ) ;
 copy( &BinDx , &BinAux ) ;
 createLi( &BinXMin , tailleBin ) ;
 BigFloat_deciToBigFloat( &Min , &BinXMin ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 create_BigFloat_deci2( &Min , taille , yInf ) ;
 create_BigFloat_deci2( &Max , taille , ySup ) ;
 sub_deci( &d , &Min , &Max ) ;
 createLi( &BinDy , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDy ) ;
 div_by_pow2_bin( &BinAux , &BinDy , powerMatrix ) ;
 copy( &BinDy , &BinAux ) ;
 createLi( &BinYMax , tailleBin ) ;
 BigFloat_deciToBigFloat( &Max , &BinYMax ) ;

 createLi( &BinY , tailleBin ) ;
 createLi( &BinCx , tailleBin ) ;
 createLi( &BinCy , tailleBin ) ;
 createLi( &BinRe , tailleBin ) ;
 createLi( &BinIm , tailleBin ) ;
 createLi( &a0 , tailleBin ) ;
 createLi( &a1 , tailleBin ) ;
 createLi( &a2 , tailleBin ) ;
 createLi( &prod , 2*tailleBin+2 ) ; // pas de souci ;)

 printf("\n") ;

 for( i = 0 ; i < sizeMatrix ; i++ )
        {
	 mulB_bin( &BinAux  , &BinDy , i ) ;
         add_bin( &BinY , &BinYMax , &BinAux ) ;
         for( j = 0 ; j < sizeMatrix ; j++ )
                {
		 mulB_bin( &BinAux , &BinDx , j ) ;
                 add_bin( &BinRe , &BinXMin , &BinAux ) ;
		 copy( &BinIm , &BinY ) ;
                 copy( &BinCx , &BinRe ) ;
                 copy( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ,
		 						    &a0 , &a1 , &a2 , &prod , tailleBin) ;

                }
        }

 destroyLi( &a0) ;
 destroyLi( &a1) ;
 destroyLi( &a2) ;
 destroyLi( &prod) ;
 destroyLi( &BinRe ) ;
 destroyLi( &BinIm ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 destroyLi( &d ) ;
 destroyLi( &BinCx ) ;
 destroyLi( &BinCy ) ;
 destroyLi( &BinYMax ) ;
 destroyLi( &BinXMin ) ;
 destroyLi( &BinY ) ;
 destroyLi( &BinDx) ;
 destroyLi( &BinDy ) ;
 destroyLi( &BinAux ) ;

 return 0 ;
}


int calcul_Mandelbrot_1Quart( char* xInf ,char* xSup ,char* yInf ,char* ySup ,
		              long tailleBin,long iter ,long **hypert , long corner )
{
 Li Min ;
 Li Max ;
 Li d   ;
 long i, j , taille ;
 Li  BinRe  ;
 Li  BinIm  ;
 Li  BinCx  ;
 Li  BinCy  ;
 Li  BinDx  ;
 Li  BinDy  ;
 Li  BinAux ;
 Li  BinY   ;
 Li  BinX   ; // pas vraiment utile !
 Li BinYMax ;
 Li BinXMin ;
 Li a0 ;
 Li a1 ;
 Li a2 ;
 Li prod ;
 //time_t tops ,tops2 ;

 taille = (long int) ceil( 1.00343331888 * tailleBin ) + 8 ;
 create_BigFloat_deci2( &Min , taille , xInf ) ;
 create_BigFloat_deci2( &Max , taille , xSup ) ;
 createLi( &d , taille ) ;
 sub_deci( &d , &Max , &Min ) ;
 createLi( &BinDx , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDx ) ;
 createLi( &BinAux , tailleBin ) ;
 div_by_pow2_bin( &BinAux , &BinDx , 9 ) ;
 copy( &BinDx , &BinAux ) ;
 createLi( &BinXMin , tailleBin ) ;
 BigFloat_deciToBigFloat( &Min , &BinXMin ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 create_BigFloat_deci2( &Min , taille , yInf ) ;
 create_BigFloat_deci2( &Max , taille , ySup ) ;
 sub_deci( &d , &Min , &Max ) ;
 createLi( &BinDy , tailleBin ) ;
 BigFloat_deciToBigFloat( &d , &BinDy ) ;
 div_by_pow2_bin( &BinAux , &BinDy , 9 ) ;
 copy( &BinDy , &BinAux ) ;
 createLi( &BinYMax , tailleBin ) ;
 BigFloat_deciToBigFloat( &Max , &BinYMax ) ;

 createLi( &BinY , tailleBin ) ;
 createLi( &BinX , tailleBin ) ;
 createLi( &BinCx , tailleBin ) ;
 createLi( &BinCy , tailleBin ) ;
 createLi( &BinRe , tailleBin ) ;
 createLi( &BinIm , tailleBin ) ;
 createLi( &a0 , tailleBin ) ;
 createLi( &a1 , tailleBin ) ;
 createLi( &a2 , tailleBin ) ;
 createLi( &prod , 2*tailleBin+2 ) ; // pas de souci ;)

 //tops = clock() ;
 //tops = time(NULL) ;

 // je fais les lignes suivantes 8  8 ...
 for( i = 7+8*corner ; i < 512-8*corner ; i += 8 )
        {
	 mulB_bin( &BinAux  , &BinDy , i ) ;
         add_bin( &BinY , &BinYMax , &BinAux ) ;
         for( j = 8*corner ; j < 512-8*corner ; j++ )
                {
		 mulB_bin( &BinAux , &BinDx , j ) ;
                 add_bin( &BinRe , &BinXMin , &BinAux ) ;
		 copy( &BinIm , &BinY ) ;
                 copy( &BinCx , &BinRe ) ;
                 copy( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter,
		 						    &a0 , &a1 , &a2 , &prod , tailleBin ) ;
                }
        }

 // je fais les colonnes suivantes 8  8 ...
 for( j = 7+8*corner ; j < 512-8*corner ; j += 8 )
        {
	 mulB_bin( &BinAux , &BinDx , j ) ;
	 add_bin( &BinX , &BinXMin , &BinAux ) ;
	 for( i = 1+8*corner ; i < 512-8*corner ; i++ )
	 	{
		 if ( i % 8 != 7 ) // pour ne pas recalculer les pixels dj faites prcdemment.
		 	{
			 mulB_bin( &BinAux  , &BinDy , i ) ;
         	 	 add_bin( &BinIm , &BinYMax , &BinAux ) ;
                 	 copy( &BinRe , &BinX ) ;
		 	 copy( &BinCy , &BinIm ) ;
			 copy( &BinCx , &BinX ) ;
			 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter,
			 						    &a0 , &a1 , &a2 , &prod , tailleBin ) ;
			}
		}
	}


// tops2 = clock() ;
 //tops2 = time(NULL) ;
 //printf(" Dure du calcul = %10.3f secondes.\n" , (double) (tops2-tops)/ CLOCKS_PER_SEC ) ;
 //printf(" Dure du calcul = %10.3f secondes.\n" , difftime(tops2,tops) ) ;


 destroyLi( &a0) ;
 destroyLi( &a1) ;
 destroyLi( &a2) ;
 destroyLi( &prod) ;
 destroyLi( &BinRe ) ;
 destroyLi( &BinIm ) ;
 destroyLi( &Min ) ;
 destroyLi( &Max ) ;
 destroyLi( &BinCx ) ;
 destroyLi( &BinCy ) ;
 destroyLi( &BinYMax ) ;
 destroyLi( &BinXMin ) ;
 destroyLi( &BinY ) ;
 destroyLi( &BinX ) ;
 destroyLi( &BinDx) ;
 destroyLi( &BinDy ) ;
 destroyLi( &BinAux ) ;

 return 0 ;
}

/*
int utilisation_Mandelbrot_9( char* xInf ,
                              char* xSup ,
                              char* yInf ,
                              char* ySup ,
                              long iter ,
                              char* name  )
{
 Li Min ;
 Li Max ;
 Li d   ;
 long i, j ;
 Li  BinRe  ;
 Li  BinIm  ;
 Li  BinCx  ;
 Li  BinCy  ;
 Li  BinDx  ;
 Li  BinDy  ;
 Li  BinAux ;
//Li_9  BinX  ;
 Li  BinY   ;
 Li BinYMax ;
 Li BinXMin ;

 int lecture_reel ;
 clock_t tops ,tops2 ;
 double profondeurX , profondeurY , profondeur ;

 long **hypert ;
 FILE* fil ;
 long buffer[512] ;

 lecture_reel  = create_BigFloat_deci_12( &Min , xInf ) ;
 lecture_reel += create_BigFloat_deci_12( &Max , xSup ) ;
 if ( lecture_reel != 0 )
 	{
	 printf(" Problem !!!") ;
	 return 1 ;
	}
 sub_deci_12( &d , &Max , &Min ) ;
 BigFloat_deci_12toBigFloat_9( &d , &BinDx ) ;
 profondeurX = depth_12( &d ) ;
 BigFloat_deci_12toBigFloat_9( &Min , &BinXMin ) ;

 lecture_reel += create_BigFloat_deci_12( &Min , yInf ) ;
 lecture_reel += create_BigFloat_deci_12( &Max , ySup ) ;
 if ( lecture_reel != 0 )
 	{
	 printf(" Problem !!!") ;
	 return 1 ;
	}
 sub_deci_12( &d , &Min , &Max ) ;
 BigFloat_deci_12toBigFloat_9( &d , &BinDy ) ;
 profondeurY = depth_12( &d ) ;
 BigFloat_deci_12toBigFloat_9( &Max , &BinYMax ) ;
 if ( profondeurX < profondeurY ) profondeur = profondeurX ;
                                  else
                                  profondeur = profondeurY ;
 printf(" La profondeur est %3.1f \n" , profondeur ) ;

 div_by_pow2_bin_9( &BinAux , &BinDx , 9 ) ;
 copy_9( &BinDx , &BinAux ) ;
 div_by_pow2_bin_9( &BinAux , &BinDy , 9 ) ;
 copy_9( &BinDy , &BinAux ) ;

 hypert = CreateMatrix( 512 ) ;

 tops = times(NULL) ;

 for( i = 0 ; i < 512 ; i++ )
        {
	 mulB_bin_9( &BinAux  , &BinDy , i ) ;
         add_bin_9( &BinY , &BinYMax , &BinAux ) ;
         for( j = 0 ; j < 512 ; j++ )
                {
		 mulB_bin_9( &BinAux , &BinDx , j ) ;
                 add_bin_9( &BinRe , &BinXMin , &BinAux ) ;
		 copy_9( &BinIm , &BinY ) ;
                 copy_9( &BinCx , &BinRe ) ;
                 copy_9( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
                }
        }

 tops2 = times(NULL) ;
 printf(" Dure du calcul = %10.3f secondes.\n" , (tops2-tops)/ 100.0 ) ;

  fil =fopen( name , "w" ) ;
 for( i = 0 ; i < 512 ; i++ )
        {
         for( j = 0 ; j < 512 ; j++ )
                 *( buffer + j ) = *( *( hypert + i) + j ) ;
         fwrite( buffer , sizeof(long) , 512 , fil ) ;
        }
 fclose( fil ) ;

 destroyLiMatrix( hypert , 512 ) ;

 return 0 ;
}
*/
/*
int utilisation_Mandelbrot( char* xInf ,char* xSup ,char* yInf ,char* ySup ,long iter ,char* name  )
{
 Li_12 Min ;
 Li_12 Max ;
 Li_12 d   ;
 long i, j , iDeb , jDeb , iloc , jloc , iCel , jCel , nbCellule , iPas , jPas ;
 unsigned long valeur_iter , calcul_cellule , nb_Cel_NonCal = 0 ;
 Li_9 BinRe  ;
 Li_9 BinIm  ;
 Li_9 BinCx  ;
 Li_9 BinCy  ;
 Li_9 BinDx  ;
 Li_9 BinDy  ;
 Li_9 BinAux ;
//Li_9  BinX  ;
 Li_9 BinY   ;
 Li_9 BinX   ;
 Li_9 BinYMax ;
 Li_9 BinXMin ;

 int lecture_reel ;
 clock_t tops ,tops2 ;
 double profondeurX , profondeurY , profondeur ;

 long **hypert ;
 FILE* fil ;
 long buffer[512] ;

 lecture_reel  = create_BigFloat_deci_12( &Min , xInf ) ;
 lecture_reel += create_BigFloat_deci_12( &Max , xSup ) ;
 if ( lecture_reel != 0 )
 	{
	 printf(" Problem !!!") ;
	 return 1 ;
	}
 sub_deci_12( &d , &Max , &Min ) ;
 BigFloat_deci_12toBigFloat_9( &d , &BinDx ) ;
 profondeurX = depth_12( &d ) ;
 BigFloat_deci_12toBigFloat_9( &Min , &BinXMin ) ;

 lecture_reel += create_BigFloat_deci_12( &Min , yInf ) ;
 lecture_reel += create_BigFloat_deci_12( &Max , ySup ) ;
 if ( lecture_reel != 0 )
 	{
	 printf(" Problem !!!") ;
	 return 1 ;
	}
 sub_deci_12( &d , &Min , &Max ) ;
 BigFloat_deci_12toBigFloat_9( &d , &BinDy ) ;
 profondeurY = depth_12( &d ) ;
 BigFloat_deci_12toBigFloat_9( &Max , &BinYMax ) ;
 if ( profondeurX < profondeurY ) profondeur = profondeurX ;
                                  else
                                  profondeur = profondeurY ;
 printf(" La profondeur est %3.1f \n" , profondeur ) ;

 div_by_pow2_bin_9( &BinAux , &BinDx , 9 ) ;
 copy_9( &BinDx , &BinAux ) ;
 div_by_pow2_bin_9( &BinAux , &BinDy , 9 ) ;
 copy_9( &BinDy , &BinAux ) ;

 hypert = CreateMatrix( 512 ) ;

 tops = times(NULL) ;

 // premire ligne en haut de l'cran.
 copy_9( &BinCy , &BinYMax ) ;
 for( j = 0 ; j < 512 ; j++ )
	{
	 mulB_bin_9( &BinAux , &BinDx , j ) ;
	 add_bin_9( &BinRe , &BinXMin , &BinAux ) ;
	 copy_9( &BinIm , &BinYMax ) ;
	 copy_9( &BinCx , &BinRe ) ;
	 *( *( hypert ) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
	}
 // je fais les lignes suivantes 8  8 ...
 for( i = 7 ; i < 512 ; i += 8 )
        {
	 mulB_bin_9( &BinAux  , &BinDy , i ) ;
         add_bin_9( &BinY , &BinYMax , &BinAux ) ;
         for( j = 0 ; j < 512 ; j++ )
                {
		 mulB_bin_9( &BinAux , &BinDx , j ) ;
                 add_bin_9( &BinRe , &BinXMin , &BinAux ) ;
		 copy_9( &BinIm , &BinY ) ;
                 copy_9( &BinCx , &BinRe ) ;
                 copy_9( &BinCy , &BinY ) ;
		 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
                }
        }
 // Je m'occupe maintenant de la premire colonne  gauche.
 copy_9( &BinCx , &BinXMin ) ;
 for( i = 1 ; i < 512 ; i++ ) // je pars de  i = 1 car j'ai dj calculer la premire ligne.
        {
	 if ( i % 8 != 7 ) // pour ne pas recalculer les pixels dj faites prcdemment.
	 	{
		 mulB_bin_9( &BinAux  , &BinDy , i ) ;
         	 add_bin_9( &BinIm , &BinYMax , &BinAux ) ;
                 copy_9( &BinRe , &BinXMin ) ;
		 copy_9( &BinCy , &BinIm ) ;
		 *( *( hypert + i) ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
		}
	}
 // je fais les colonnes suivantes 8  8 ...
 for( j = 7 ; j < 512 ; j += 8 )
        {
	 mulB_bin_9( &BinAux , &BinDx , j ) ;
	 add_bin_9( &BinX , &BinXMin , &BinAux ) ;
	 for( i = 1 ; i < 512 ; i++ )
	 	{
		 if ( i % 8 != 7 ) // pour ne pas recalculer les pixels dj faites prcdemment.
		 	{
			 mulB_bin_9( &BinAux  , &BinDy , i ) ;
         	 	 add_bin_9( &BinIm , &BinYMax , &BinAux ) ;
                 	 copy_9( &BinRe , &BinX ) ;
		 	 copy_9( &BinCy , &BinIm ) ;
			 copy_9( &BinCx , &BinX ) ;
			 *( *( hypert + i) + j ) = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
			}
		}
	}
// A ce niveau, le quadrillge est calcul. Passons  l'intrieur des cellules.

 nbCellule = 512 / 8 ;
 for( iCel = 0 ; iCel < nbCellule ; iCel++ )
 	{
	 if ( iCel == 0 )
	 	{
		 iDeb = 0 ;
		 iPas = 7 ;
		}
		else
		{
		 iDeb = iCel * 8 - 1 ;
		 iPas = 8 ;
		}
	 for( jCel = 0 ; jCel < nbCellule ; jCel++ )
		{
		 // Je teste si le bord de la cellule a des pixels tous  la valeur valeur_iter.
		 if ( jCel == 0 )
		 	{
			 jDeb = 0 ;
			 jPas = 7 ;
			}
			else
			{
			 jDeb = jCel * 8 - 1 ;
			 jPas = 8 ;
			}
		 valeur_iter = *( *( hypert + iDeb ) + jDeb ) ;
		 calcul_cellule = 0 ;
		 for( jloc = jDeb ; jloc <= jDeb + jPas ; jloc++ )
		 	if ( *( *( hypert + iDeb ) + jloc ) != valeur_iter )
				{
				 calcul_cellule = 1 ;
				 break ;
				}
		 if ( calcul_cellule == 0 )
		 	{
			 for( jloc = jDeb ; jloc <= jDeb + jPas ; jloc++ )
				if ( *( *( hypert + iDeb+iPas ) + jloc ) != valeur_iter )
					{
					 calcul_cellule = 1 ;
					 break ;
					}
			}
		 if ( calcul_cellule == 0 )
		 	for( iloc = iDeb+1 ; iloc < iDeb + iPas ; iloc++ )
				if ( *( *( hypert + iloc ) + jDeb ) != valeur_iter )
					{
					 calcul_cellule = 1 ;
					 break ;
					}
		 if ( calcul_cellule == 0 )
		 	for( iloc = iDeb+1 ; iloc < iDeb + iPas ; iloc++ )
				if ( *( *( hypert + iloc ) + (jDeb+jPas) ) != valeur_iter )
					{
					 calcul_cellule = 1 ;
					 break ;
					}

		if ( calcul_cellule == 1 )
			 	{
				 for( iloc = iDeb+1 ; iloc < iDeb + iPas ; iloc++ )
				 	{
					 mulB_bin_9( &BinAux  , &BinDy , iloc ) ;
      					 add_bin_9( &BinCy , &BinYMax , &BinAux ) ;
					 //copy_9( &BinCy , &BinY ) ;
					 for( jloc = jDeb+1 ; jloc < jDeb + jPas ; jloc++ )
						{
						 mulB_bin_9( &BinAux , &BinDx , jloc ) ;
						 add_bin_9( &BinRe , &BinXMin , &BinAux ) ;
						 copy_9( &BinIm , &BinCy ) ;
						 copy_9( &BinCx , &BinRe ) ;
						 *( *( hypert + iloc ) + jloc )  = calcul_germe_mandelbrot( &BinRe , &BinIm , &BinCx , &BinCy , iter ) ;
						}
					}
				 //printf("+") ;
				}
				else
				{
				 for( iloc = iDeb+1 ; iloc < iDeb + iPas ; iloc++ )
				 	for( jloc = jDeb+1 ; jloc < jDeb + jPas ; jloc++ )
						*( *( hypert + iloc ) + jloc )  = valeur_iter ;
				 nb_Cel_NonCal++ ; //printf("-(%ld,%ld)" , jCel,iCel) ;
				}
		}
	}

 tops2 = times(NULL) ;
 printf(" Dure du calcul = %10.3f secondes.\n" , (tops2-tops)/ 100.0 ) ;
 printf(" Poucentage de points non calculs = %3.2f \n" , nb_Cel_NonCal * 100.0 / nbCellule / nbCellule ) ;
  fil =fopen( name , "w" ) ;
 for( i = 0 ; i < 512 ; i++ )
        {
         for( j = 0 ; j < 512 ; j++ )
                 *( buffer + j ) = *( *( hypert + i) + j ) ;
         fwrite( buffer , sizeof(long) , 512 , fil ) ;
        }
 fclose( fil ) ;

 DestroyMatrix( hypert , 512 ) ;

 return 0 ;
}
*/


