/*
 * Routine conu par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include "lf32.h"
#include "op_fftnr.h"
#include "fft_common.h"


// cette routine calcule cos(l*pi/2^x) et sin(l*pi/2^x)
// on doit avoir x et l positifs,et
void cossin( long double* co, long double* si, long l, long x )
{
 long double a,b,t ;
 long k,i ;

 if ( (l & PARITE) == 0 )
	{
	 a = 1.0 ;
	 b = 0.0 ;
	}
	else
	{
	 a = cosinusl[x] ;
	 b = sinusl[x] ;
	}
 l >>= 1 ;
 k = 0 ;
 while (( k<x) && (l>0))
	{
	 if ( (l & PARITE) == 1 )
		{
		 i = x-k-1 ;
		 t = a * cosinusl[i] - b * sinusl[i] ;
		 b = b * cosinusl[i] + a * sinusl[i] ;
		 a = t ;
		}
	 l >>= 1 ;
	 k++ ;
	}
 if ( (l & PARITE) == 1 )
	{
	 *co = -a ;
	 *si = -b ;
	}
	else
	{
	 *co = a ;
	 *si = b ;
	}
}

void four1( double* data, unsigned long nn, int isign )
{
 unsigned long n,mmax,m,j,istep,i,expoo ;
 double wtemp,wr,wi,wpr,wpi,tempr,tempi ;

 n = nn << 1 ;
 j = 1 ;
 for( i=1 ; i<n ; i+=2 )
	{
	 if (j > i )
		{
		 tempr = data[j] ;
		 data[j] = data[i] ;
		 data[i] = tempr ;
		 tempr = data[j+1] ;
		 data[j+1] = data[i+1] ;
		 data[i+1] = tempr ;
		}
	 m = nn ;
	 while ((m >= 2) && (j > m))
		{
		 j -= m ;
		 m >>= 1 ;
		}
	 j += m ;
	}

 mmax = 2 ;
 expoo = 1 ;
 while( n > mmax )
	{
	 istep = mmax << 1 ;
	 //theta = isign * (6.28318530717959/mmax) ;
	 //wtemp = sin( 0.5*theta ) ;
	 wtemp = sinus[expoo] ;
	 wtemp = wtemp*wtemp ;
	 wpr = - wtemp - wtemp ;
	 //wpi = sin( theta ) ;
	 wpi = isign * sinus[expoo-1] ;
	 wr = 1.0 ;
	 wi = 0.0 ;
	 for( m=1 ; m<mmax ; m+=2 )
		{
		 for( i=m ; i <=n ; i+=istep )
			{
			 j = i + mmax ;
			 tempr = wr * data[j]   - wi * data[j+1] ;
			 tempi = wr * data[j+1] + wi * data[j] ;
			 data[j] = data[i] - tempr ;
			 data[j+1] = data[i+1] - tempi ;
			 data[i] += tempr ;
			 data[i+1] += tempi ;
			}
		 wtemp = wr ;
		 wr = wtemp * wpr - wi * wpi + wr ;
		 wi = wi * wpr + wtemp * wpi + wi ;
		}
	 mmax = istep ;
	 expoo++ ;
	}
}

void four1l( long double* data, unsigned long nn, int isign )
{
 long double wtemp,wr,wi,wpr,wpi,tempr,tempi,lisign ;
 unsigned long n,mmax,m,j,istep,i,expoo ;

 if ( isign == 1 ) lisign = 1.0 ; else lisign = -1.0 ;
 n = nn << 1 ;
 j = 1 ;
 for( i=1 ; i<n ; i+=2 )
	{
	 if (j > i )
		{
		 tempr = data[j] ;
		 data[j] = data[i] ;
		 data[i] = tempr ;
		 tempr = data[j+1] ;
		 data[j+1] = data[i+1] ;
		 data[i+1] = tempr ;
		}
	 m = nn ;
	 while ((m >= 2) && (j > m))
		{
		 j -= m ;
		 m >>= 1 ;
		}
	 j += m ;
	}

 mmax = 2 ;
 expoo = 1 ;
 while( n > mmax )
	{
	 istep = mmax << 1 ;
	 //theta = isign * (6.28318530717959/mmax) ;
	 //wtemp = sin( 0.5*theta ) ;
	 wtemp = sinusl[expoo] ;
	 wtemp = wtemp*wtemp ;
	 wpr = - wtemp - wtemp ;
	 //wpi = sin( theta ) ;
	 wpi = isign * sinusl[expoo-1] ;
	 wr = 1.0 ;
	 wi = 0.0 ;
	 for( m=1 ; m<mmax ; m+=2 )
		{
		 for( i=m ; i <=n ; i+=istep )
			{
			 j = i + mmax ;
			 tempr = wr * data[j]   - wi * data[j+1] ;
			 tempi = wr * data[j+1] + wi * data[j] ;
			 data[j] = data[i] - tempr ;
			 data[j+1] = data[i+1] - tempi ;
			 data[i] += tempr ;
			 data[i+1] += tempi ;
			}
		 wtemp = wr ;
		 wr = wtemp * wpr - wi * wpi + wr ;
		 wi = wi * wpr + wtemp * wpi + wi ;
		}
	 mmax = istep ;
	 expoo++ ;
	}
}

// Cette routine permet l'initialisation des varaibles ncessaires  la
// multiplication de deux entiers de type Li de taille 2^k en base 2^30.
void init_mul_fftnr( long k, double** inoutA , double** inoutB )
{
 long tailleFFT ;

 // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.

 tailleFFT = 1 << (k+1) ;

 *inoutA  = malloc( 2*tailleFFT * sizeof(double) ) ;
 *inoutB  = malloc( 2*tailleFFT * sizeof(double) ) ;
}

// Cette routine permet l'initialisation des varaibles ncessaires  la
// multiplication de deux entiers de type Li de taille 2^k en base 2^30.
void init_mul_fftnrl( long k, long double** inoutA , long double** inoutB )
{
 long tailleFFT ;


 // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.
 tailleFFT = 1 << (k+1) ;

 *inoutA  = malloc( 2*tailleFFT * sizeof(long double) ) ;
 *inoutB  = malloc( 2*tailleFFT * sizeof(long double) ) ;
}

// Cette routine permet l'initialisation des varaibles ncessaires  la
// multiplication de deux entiers de type Li de taille 2^k en base 2^30.
void init_square_fftnr( long k, double** inoutA )
{
 long tailleFFT ;

 // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.

 tailleFFT = 1 << (k+1) ;

 *inoutA  = malloc( 2*tailleFFT * sizeof(double) ) ;
}

void init_square_fftnrl( long k, long double** inoutA )
{
 long tailleFFT ;

 // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.

 tailleFFT = 1 << (k+1) ;

 *inoutA  = malloc( 2*tailleFFT * sizeof(long double) ) ;
}

void destroy_mul_fftnr( double* inoutA , double* inoutB )
{
 free(inoutA) ;
 free(inoutB) ;
}

void destroy_mul_fftnrl( long double* inoutA , long double* inoutB )
{
 free(inoutA) ;
 free(inoutB) ;
}

void destroy_square_fftnr( double* inoutA )
{
 free(inoutA) ;
}

void destroy_square_fftnrl( long double* inoutA )
{
 free(inoutA) ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
double mul_2P_fftnr_bin_test( Li* prod , Li* a , Li* b , long k ,
		     double* inoutA , double* inoutB )
{
 double error=0.0 ;
 double errorUnite = 0.0 ;
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1( inoutA-1,taille,-1 ) ;
 four1( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

// for(i=1;i<htaille;i++)
 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 //i2 = -(k+1) ;
 //for(i=0;i<taille;i++)
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floor( tmp1 + 0.5 ) ;
	 tmp4 = floor( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 //printf("\n errorUnite = %1.4lf" , errorUnite ) ;

 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void mul_2P_fftnr_bin( Li* prod , Li* a , Li* b , long k ,
		     double* inoutA , double* inoutB )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1( inoutA-1,taille,-1 ) ;
 four1( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 ll1 = floor( inoutA[in] * i_taille + 0.5 ) ;
	 ll2 = floor( inoutA[in+1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}


// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
double square_2P_fftnr_bin_test( Li* prod , Li* a , long k ,
		     double* inoutA )
{
 double error=0.0 ;
 double errorUnite = 0.0 ;
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Ia,RaM,IaM ;
 double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}


 four1( inoutA-1,taille,-1 ) ;

 // ------  i = 0
 tmp0 = inoutA[0] + inoutA[1] ;
 tmp1 = inoutA[0] - inoutA[1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[taille] ;
 tmp4 = inoutA[taille+1] ;
 tmp0 = (tmp3-tmp4)*(tmp3+tmp4) ;
 tmp1 = 2.0 * tmp3 * tmp4 ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floor( tmp1 + 0.5 ) ;
	 tmp4 = floor( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 printf("\n errorUnite = %1.lg" , errorUnite ) ;

 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void square_2P_fftnr_bin( Li* prod , Li* a , long k ,
		     double* inoutA )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Ia,RaM,IaM ;
 double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}


 four1( inoutA-1,taille,-1 ) ;

 // ------  i = 0
 tmp0 = inoutA[0] + inoutA[1] ;
 tmp1 = inoutA[0] - inoutA[1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[taille] ;
 tmp4 = inoutA[taille+1] ;
 tmp0 = (tmp3-tmp4)*(tmp3+tmp4) ;
 tmp1 = 2.0 * tmp3 * tmp4 ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 ll1 = floor( inoutA[in] * i_taille + 0.5 ) ;
	 ll2 = floor( inoutA[in+1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
long double mul_2P_fftnrl_bin_test( Li* prod , Li* a , Li* b , long k ,
		     long double* inoutA , long double* inoutB )
{
 long double error=0.0 ;
 long double errorUnite = 0.0 ;
 long double tmp0 ;
 long double tmp1 ;
 long double tmp2 ;
 long double tmp3 ;
 long double tmp4 ;
 long double tmp5 ;
 long double tmp6 ;
 long double tmp7 ;
 long double tmp8 ;
 long double tmp9 ;
 long double tmp10 ;
 long double i_taille ;
 long double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 long double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1l( inoutA-1,taille,-1 ) ;
 four1l( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinusl[k+1] ;
 tmp10 = win ;
 tmp9 = sinusl[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1l( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floorl( tmp1 + 0.5 ) ;
	 tmp4 = floorl( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 printf("\n errorUnite = %1.Lg" , errorUnite ) ;

 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void mul_2P_fftnrl_bin( Li* prod , Li* a , Li* b , long k ,
		     long double* inoutA , long double* inoutB )
{
 long double tmp0 ;
 long double tmp1 ;
 long double tmp2 ;
 long double tmp3 ;
 long double tmp4 ;
 long double tmp5 ;
 long double tmp6 ;
 long double tmp7 ;
 long double tmp8 ;
 long double tmp9 ;
 long double tmp10 ;
 long double i_taille ;
 long double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 long double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (long double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1l( inoutA-1,taille,-1 ) ;
 four1l( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinusl[k+1] ;
 tmp10 = win ;
 tmp9 = sinusl[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1l( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 ll1 = floorl( inoutA[in] * i_taille + 0.5 ) ;
	 ll2 = floorl( inoutA[in+1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}


// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
long double square_2P_fftnrl_bin_test( Li* prod , Li* a , long k ,
		     long double* inoutA )
{
 long double error=0.0 ;
 long double errorUnite = 0.0 ;
 long double tmp0 ;
 long double tmp1 ;
 long double tmp2 ;
 long double tmp3 ;
 long double tmp4 ;
 long double tmp5 ;
 long double tmp6 ;
 long double tmp7 ;
 long double tmp8 ;
 long double tmp9 ;
 long double tmp10 ;
 long double i_taille ;
 long double Ra,Ia,RaM,IaM ;
 long double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (long double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}


 four1l( inoutA-1,taille,-1 ) ;

 // ------  i = 0
 tmp0 = inoutA[0] + inoutA[1] ;
 tmp1 = inoutA[0] - inoutA[1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinusl[k+1] ;
 tmp10 = win ;
 tmp9 = sinusl[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[taille] ;
 tmp4 = inoutA[taille+1] ;
 tmp0 = (tmp3-tmp4)*(tmp3+tmp4) ;
 tmp1 = 2.0 * tmp3 * tmp4 ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1l( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floorl( tmp1 + 0.5 ) ;
	 tmp4 = floorl( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 printf("\n errorUnite = %1.Lg" , errorUnite ) ;

 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void square_2P_fftnrl_bin( Li* prod , Li* a , long k , long double* inoutA )
{
 long double tmp0 ;
 long double tmp1 ;
 long double tmp2 ;
 long double tmp3 ;
 long double tmp4 ;
 long double tmp5 ;
 long double tmp6 ;
 long double tmp7 ;
 long double tmp8 ;
 long double tmp9 ;
 long double tmp10 ;
 long double i_taille ;
 long double Ra,Ia,RaM,IaM ;
 long double win,wrn ;
 long i , t , t2 ,in , inM ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (long double) 1.0 / taille ;

 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}


 four1l( inoutA-1,taille,-1 ) ;

 // ------  i = 0
 tmp0 = inoutA[0] + inoutA[1] ;
 tmp1 = inoutA[0] - inoutA[1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinusl[k+1] ;
 tmp10 = win ;
 tmp9 = sinusl[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[taille] ;
 tmp4 = inoutA[taille+1] ;
 tmp0 = (tmp3-tmp4)*(tmp3+tmp4) ;
 tmp1 = 2.0 * tmp3 * tmp4 ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1l( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 ll1 = floorl( inoutA[in] * i_taille + 0.5 ) ;
	 ll2 = floorl( inoutA[in+1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}





// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
double mul_2P_fftnr_bini_test( Lii* prod , Lii* a , Lii* b , long k ,
		     double* inoutA , double* inoutB )
{
 double error=0.0 ;
 double errorUnite = 0.0 ;
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 long i , t , t2 ,in , inM , indiceMax ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 indiceMax = a->indiceMax ;
 for(i=0,in=0;i<=indiceMax;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=(indiceMax+1)<<1;in<taille;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 indiceMax = b->indiceMax ;
 for(i=0,in=0;in<taille;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=(indiceMax+1)<<1;in<taille;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1( inoutA-1,taille,-1 ) ;
 four1( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

// for(i=1;i<htaille;i++)
 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 //i2 = -(k+1) ;
 //for(i=0;i<taille;i++)
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floor( tmp1 + 0.5 ) ;
	 tmp4 = floor( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 //printf("\n errorUnite = %1.4lf" , errorUnite ) ;
 i = a->indiceMax + b->indiceMax + 1 ;
 while (( i >= 0 ) && ( prod->mantisse[i] == 0 )) i-- ;
 prod->indiceMax = i ;
 if (( i <= 0 ) && ( prod->mantisse[0] == 0 ))
 	prod->signe = 0 ; else prod->signe = 1 ;

 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void mul_2P_fftnr_bini( Lii* prod , Lii* a , Lii* b , long k ,
		     double* inoutA , double* inoutB )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double i_taille ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 long i , t , t2 ,in , inM , indiceMax ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (double) 1.0 / taille ;

 indiceMax = a->indiceMax ;
 //prod->mantisse[indiceMax+b->indiceMax+1] = 0 ;
 for(i=0,in=0;i<=indiceMax;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=(indiceMax+1)<<1;in<taille;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 indiceMax = b->indiceMax ;
 for(i=0,in=0;i<=indiceMax;i++,in+=2)
	{
	 t = b->mantisse[i] ;
	 inoutB[in] = t & HALF_CACHE32_BIN ;
	 inoutB[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=(indiceMax+1)<<1;in<taille;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutB[in] = 0.0 ;
	 inoutB[in+1] = 0.0 ;
	}

 four1( inoutA-1,taille,-1 ) ;
 four1( inoutB-1,taille,-1 ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0] + inoutA[1] ) * ( inoutB[0] + inoutB[1] ) ;
 tmp2 = ( inoutA[0] - inoutA[1] ) * ( inoutB[0] - inoutB[1] ) ;
 inoutA[0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[in] + inoutB[inM] ) ;
	 tmp2 = 0.5 * ( inoutB[in] - inoutB[inM] ) ;
	 tmp3 = 0.5 * ( inoutB[in+1] + inoutB[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutB[in+1] - inoutB[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[taille] * inoutB[taille] - inoutA[taille+1] * inoutB[taille+1] ;
 tmp1 = inoutA[taille] * inoutB[taille+1] + inoutA[taille+1] * inoutB[taille] ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 ll1 = floor( inoutA[in] * i_taille + 0.5 ) ;
	 ll2 = floor( inoutA[in+1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}

 i = a->indiceMax + b->indiceMax + 1 ;
 while (( i >= 0 ) && ( prod->mantisse[i] == 0 )) i-- ;
 prod->indiceMax = i ;
 if (( i <= 0 ) && ( prod->mantisse[0] == 0 ))
 	prod->signe = 0 ; else prod->signe = 1 ;
}


// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
long double square_2P_fftnrl_bini_test( Lii* prod , Lii* a , long k ,
		     long double* inoutA )
{
 long double error=0.0 ;
 long double errorUnite = 0.0 ;
 long double tmp0 ;
 long double tmp1 ;
 long double tmp2 ;
 long double tmp3 ;
 long double tmp4 ;
 long double tmp5 ;
 long double tmp6 ;
 long double tmp7 ;
 long double tmp8 ;
 long double tmp9 ;
 long double tmp10 ;
 long double i_taille ;
 long double Ra,Ia,RaM,IaM ;
 long double win,wrn ;
 long i , t , t2 ,in , inM , indiceMax ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;
 long taille2 = taille << 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = (long double) 1.0 / taille ;

 indiceMax = a->indiceMax ;
 for(i=0,in=0;i<=indiceMax;i++,in+=2)
	{
	 t = a->mantisse[i] ;
	 inoutA[in] = t & HALF_CACHE32_BIN ;
	 inoutA[in+1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(in=(indiceMax+1)<<1;in<taille;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}
 for(in=taille;in<taille2;in+=2)
	{
	 inoutA[in] = 0.0 ;
	 inoutA[in+1] = 0.0 ;
	}


 four1l( inoutA-1,taille,-1 ) ;

 // ------  i = 0
 tmp0 = inoutA[0] + inoutA[1] ;
 tmp1 = inoutA[0] - inoutA[1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 win = -sinusl[k+1] ;
 tmp10 = win ;
 tmp9 = sinusl[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(in=2,inM=taille2-2;in<taille;in+=2,inM-=2)
	{
	 tmp1 = 0.5 * ( inoutA[in] + inoutA[inM] ) ;
	 tmp2 = 0.5 * ( inoutA[in] - inoutA[inM] ) ;
	 tmp3 = 0.5 * ( inoutA[in+1] + inoutA[inM+1] ) ;
	 tmp4 = 0.5 * ( inoutA[in+1] - inoutA[inM+1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[in] = tmp1 + tmp2 ;
	 inoutA[in+1] = tmp3 + tmp4 ;
	 inoutA[inM] = tmp1 - tmp2 ;
	 inoutA[inM+1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[taille] ;
 tmp4 = inoutA[taille+1] ;
 tmp0 = (tmp3-tmp4)*(tmp3+tmp4) ;
 tmp1 = 2.0 * tmp3 * tmp4 ;
 inoutA[taille+1] = tmp1 ;
 inoutA[taille]   = tmp0 ;

 four1l( inoutA-1,taille,1 ) ;

 lcarry = 0 ;
 for(i=0,in=0;in<taille2;in+=2,i++)
	{
	 tmp1 = inoutA[in] * i_taille ; //tmp1 = ldexp(inoutA[2*i],i2 ) ;
	 tmp2 = inoutA[in+1] * i_taille ;
	 tmp3 = floorl( tmp1 + 0.5 ) ;
	 tmp4 = floorl( tmp2 + 0.5 ) ;
	 if ( tmp3 > tmp4) tmp0 = tmp3 ; else tmp0 = tmp4 ;
	 if ( tmp0 > errorUnite ) errorUnite = tmp0 ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = tmp1 - tmp3 ;
	 if ( tmp5 < 0 ) tmp5 = -tmp5 ;
	 tmp6 = tmp2 - tmp4 ;
	 if ( tmp6 < 0 ) tmp6 = -tmp6 ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 printf("\n errorUnite = %1.Lg" , errorUnite ) ;

 i = ((a->indiceMax) << 1) + 1 ;
 while (( i >= 0 ) && ( prod->mantisse[i] == 0 )) i-- ;
 prod->indiceMax = i ;
 if (( i <= 0 ) && ( prod->mantisse[0] == 0 ))
 	prod->signe = 0 ; else prod->signe = 1 ;

 return error ;
}







