/*
 * Routine conu par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "lf32.h"
#include "op_karatsuba.h"

void add_R_2P_karatsuba_bin( long* a , long* b , long n )
{
 long i , en , carry ;

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a[i] + b[i] + carry ;
	 carry = en >> DECALAGE32_BIN ;
	 a[i] = en & CACHE32_BIN ; // cache = 2^30 - 1
        }
 while( carry == 1 )
	{
	 en = a[i] + carry ;
	 carry = en >> DECALAGE32_BIN ;
	 a[i] = en & CACHE32_BIN ; // cache = 2^30 - 1
	 i++ ;
	}
}

long sub_R_2P_karatsuba_bin( long* a , long* b , long n )
{
 long i , en , carry ;

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a[i] - b[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         a[i] = en - carry * BASE32_BIN ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( a[en]  == 0 ) en++ ;
         a[en] = BASE32_BIN - a[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                a[i] = CACHE32_BIN - a[i] ;
         return -1 ;
        }
	else return 1 ;
}

long sub_2P_karatsuba_bin( long* r , long* a , long* b , long n )
{
 long i , en , carry ;

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a[i] - b[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         r[i] = en - carry * BASE32_BIN ;
        }
 if ( carry == -1 )
        {
         en = 0 ;
         while( r[en]  == 0 ) en++ ;
         r[en] = BASE32_BIN - r[en] ;
         en++ ;
         for( i = en ; i < n ; i++ )
                r[i] = CACHE32_BIN - r[i] ;
         return -1 ;
        }
	else return 1 ;
}

void sub_R_2P_karatsuba_infty_bin( long* a , long* b , long n )
{
 long i , en , carry ;

 carry = 0 ;
 for( i = 0 ; i < n ; i++ )
        {
         en = a[i] - b[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         a[i] = en - carry * BASE32_BIN ;
        }
 while( carry == -1 )
	{
         en = a[i] + carry ;
         carry = en >> 31 ;  // astuce pour rcuprer la retenue.
         a[i] = en - carry * BASE32_BIN ;
	 i++ ;
	}
}

void mul_karatsuba_2P_binV2( long* r , long* a , long* b , long n , long* t )
{
 long nh = n >> 1 , n2 = n << 1 , carry1 = 0 , carry2 = 0  ;
 //long i ;

 //printf(" r = %i  a = %i   b = %i n = %ld  t = %i \n  " , (r-rD)/4,(a-aD)/4,(b-bD)/4,n,(t-tD)/4 ) ;

 if ( n <= CUTTOF_MUL_KARATSUBA )
	{
	 mul_2P_naif_bin( r,a,b,n ) ;
	 //printf("--- a = %ld   b = %ld   r = %ld   n = %ld \n" , *a,*b,*r,n) ;
	 return ;
	}

 mul_karatsuba_2P_binV2( r,a,b,nh,t ) ;
 mul_karatsuba_2P_binV2( r+n,a+nh,b+nh,nh,t ) ;
 memcpy( t,r,n2*sizeof(long) ) ;
 add_R_2P_karatsuba_bin( r+nh,t,n ) ;
 add_R_2P_karatsuba_bin( r+nh,t+n,n ) ;
 memcpy( t,a,nh*sizeof(long) ) ;
 carry1 = sub_R_2P_karatsuba_bin( t,a+nh,nh ) ;
 memcpy( t+nh,b,nh*sizeof(long) ) ;
 carry2 = sub_R_2P_karatsuba_bin( t+nh,b+nh,nh ) ;
 mul_karatsuba_2P_binV2( t+n,t,t+nh,nh,t+n2 ) ;
 if ( carry1 * carry2 == 1 ) sub_R_2P_karatsuba_infty_bin( r+nh,t+n,n ) ;
 		else	add_R_2P_karatsuba_bin( r+nh,t+n,n ) ;
}

void mul_2P_karatsuba_bin( long* r , long* a , long* b , long n )
{
 long nh = n >> 1 , n2 = n << 1 , n5 = 5*n, carry1 = 0 , carry2 = 0 , tmp=0 , i ;
 unsigned long tmp2=0 ;

 if ( n <= CUTTOF_MUL_KARATSUBA )
	{
	 mul_2P_naif_bin( r,a,b,n ) ;
	 return ;
	}

 carry1 = sub_2P_karatsuba_bin( r+n5,a,a+nh,nh ) ;
 carry2 = sub_2P_karatsuba_bin( r+(n5+nh),b,b+nh,nh ) ;
 mul_2P_karatsuba_bin( r,a,b,nh ) ;
 mul_2P_karatsuba_bin( r+n,a+nh,b+nh,nh ) ;
 mul_2P_karatsuba_bin( r+n2,r+n5,r+(n5+nh),nh ) ;

 //  calcul de  a(0) * b(0) + a(1) * b(1) - (a(0) - a(1)) * (b(0) - b(1))

 tmp = 0 ;
 for(i=0;i<n;i++)
	{
	 tmp2 = r[i]+r[nh+i]+r[n+i]+tmp ;
	 tmp = tmp2 >> DECALAGE32_BIN ;
	 r[n5+i] = tmp2 & CACHE32_BIN ;
	}
 memcpy( r+nh,r+n5,n*sizeof(long) ) ;
 while( tmp > 0 )
	{
	 tmp2 = r[nh+i] + tmp ;
	 tmp = tmp2 >> DECALAGE32_BIN ;
	 r[nh+i] = tmp2 & CACHE32_BIN ; // cache = 2^30 - 1
	 i++ ;
	}

 if ( carry1 * carry2 == 1 )
	{
	 carry1 = 0 ;
	 for(i=0;i<n;i++)
	 	{
		 tmp = r[nh+i]+carry1-r[n2+i] ;
		 carry1 = tmp >> 31 ;  // astuce pour rcuprer la retenue.
		 r[nh+i] = tmp - carry1 * BASE32_BIN ;
		}
	 while( carry1 == -1 )
		{
		 tmp = r[nh+i] + carry1 ;
		 carry1 = tmp >> 31 ;  // astuce pour rcuprer la retenue.
		 r[nh+i] = tmp - carry1 * BASE32_BIN ;
		 i++ ;
		}
	}
	else
	{
	 carry1 = 0 ;
	 for(i=0;i<n;i++)
	 	{
		 tmp = r[nh+i]+carry1+r[n2+i] ;
		 carry1 = tmp >> DECALAGE32_BIN ;
		 r[nh+i] = tmp & CACHE32_BIN ; // cache = 2^30 - 1
		}
	 while( carry1 == 1 )
		{
		 tmp = r[nh+i] + carry1 ;
		 carry1 = tmp >> DECALAGE32_BIN ;
		 r[nh+i] = tmp & CACHE32_BIN ; // cache = 2^30 - 1
		 i++ ;
		}
	}
}

void square_karatsuba_2P_binV2( long* r , long* a , long n , long* t )
{
 long nh = n >> 1 , n3h = n + nh ;

 if ( n <= CUTTOF_SQUARE_KARATSUBA )
	{
	 square_2P_naif_bin( r,a,n ) ;
	 return ;
	}

 square_karatsuba_2P_binV2( r,a,nh,t ) ;
 square_karatsuba_2P_binV2( r+n,a+nh,nh,t ) ;
 memcpy( t,r,(n+n)*sizeof(long) ) ;
 add_R_2P_karatsuba_bin( r+nh,t,n ) ;
 add_R_2P_karatsuba_bin( r+nh,t+n,n ) ;
 memcpy( t,a,nh*sizeof(long) ) ;
 sub_R_2P_karatsuba_bin( t,a+nh,nh ) ;
 square_karatsuba_2P_binV2( t+nh,t,nh,t+n3h ) ;
 sub_R_2P_karatsuba_infty_bin( r+nh,t+nh,n ) ;
}


void square_2P_karatsuba_bin( long* r , long* a , long n )
{
 long nh = n >> 1 , n2 = n << 1 , n4_75 = 5*n - (nh >> 1) , carry1 = 0 , tmp=0 , i ;
 unsigned long tmp2=0 ;

 if ( n <= CUTTOF_SQUARE_KARATSUBA )
	{
	 square_2P_naif_bin( r,a,n ) ;
	 return ;
	}

 sub_2P_karatsuba_bin( r+n4_75,a,a+nh,nh ) ;
 square_2P_karatsuba_bin( r,a,nh ) ;
 square_2P_karatsuba_bin( r+n,a+nh,nh ) ;
 square_2P_karatsuba_bin( r+n2,r+n4_75,nh ) ;

 //  calcul de  a(0)^2 + a(1)^2 - (a(0) - a(1))^2

 tmp = 0 ;
 for(i=0;i<n;i++)
	{
	 tmp2 = r[i]+r[nh+i]+r[n+i]+tmp ;
	 tmp = tmp2 >> DECALAGE32_BIN ;
	 r[n4_75+i] = tmp2 & CACHE32_BIN ;
	}
 memcpy( r+nh,r+n4_75,n*sizeof(long) ) ;
 while( tmp > 0 )
	{
	 tmp2 = r[nh+i] + tmp ;
	 tmp = tmp2 >> DECALAGE32_BIN ;
	 r[nh+i] = tmp2 & CACHE32_BIN ; // cache = 2^30 - 1
	 i++ ;
	}

 carry1 = 0 ;
 for(i=0;i<n;i++)
 	{
	 tmp = r[nh+i]+carry1-r[n2+i] ;
	 carry1 = tmp >> 31 ;  // astuce pour rcuprer la retenue.
	 r[nh+i] = tmp - carry1 * BASE32_BIN ;
	}
 while( carry1 == -1 )
	{
	 tmp = r[nh+i] + carry1 ;
	 carry1 = tmp >> 31 ;  // astuce pour rcuprer la retenue.
	 r[nh+i] = tmp - carry1 * BASE32_BIN ;
	 i++ ;
	}
}



